from typing import Iterator
from typing import List
from typing import Optional
from urllib.parse import urljoin
from urllib.parse import urlparse

from jsonschema_path import SchemaPath

from openapi_core.schema.servers import is_absolute
from openapi_core.templating.datatypes import TemplateResult
from openapi_core.templating.paths.datatypes import Path
from openapi_core.templating.paths.datatypes import PathOperation
from openapi_core.templating.paths.datatypes import PathOperationServer
from openapi_core.templating.paths.exceptions import PathsNotFound
from openapi_core.templating.paths.util import template_path_len
from openapi_core.templating.util import parse
from openapi_core.templating.util import search


class SimplePathsIterator:
    def __init__(self, paths_part: str):
        self.paths_part = paths_part

    def __call__(
        self, name: str, spec: SchemaPath, base_url: Optional[str] = None
    ) -> Iterator[Path]:
        paths = spec / self.paths_part
        if not paths.exists():
            raise PathsNotFound(paths.as_uri())
        for path_name, path in list(paths.items()):
            if name == path_name:
                path_result = TemplateResult(path_name, {})
                yield Path(path, path_result)


class TemplatePathsIterator:
    def __init__(self, paths_part: str):
        self.paths_part = paths_part

    def __call__(
        self, name: str, spec: SchemaPath, base_url: Optional[str] = None
    ) -> Iterator[Path]:
        paths = spec / self.paths_part
        if not paths.exists():
            raise PathsNotFound(paths.as_uri())
        template_paths: List[Path] = []
        for path_pattern, path in list(paths.items()):
            # simple path.
            # Return right away since it is always the most concrete
            if name.endswith(path_pattern):
                path_result = TemplateResult(path_pattern, {})
                yield Path(path, path_result)
            # template path
            else:
                result = search(path_pattern, name)
                if result:
                    path_result = TemplateResult(path_pattern, result.named)
                    template_paths.append(Path(path, path_result))

        # Fewer variables -> more concrete path
        yield from sorted(template_paths, key=template_path_len)


class SimpleOperationsIterator:
    def __call__(
        self,
        method: str,
        paths_iter: Iterator[Path],
        spec: SchemaPath,
        base_url: Optional[str] = None,
    ) -> Iterator[PathOperation]:
        for path, path_result in paths_iter:
            if method not in path:
                continue
            operation = path / method
            yield PathOperation(path, operation, path_result)


class CatchAllMethodOperationsIterator(SimpleOperationsIterator):
    def __init__(self, ca_method_name: str, ca_operation_name: str):
        self.ca_method_name = ca_method_name
        self.ca_operation_name = ca_operation_name

    def __call__(
        self,
        method: str,
        paths_iter: Iterator[Path],
        spec: SchemaPath,
        base_url: Optional[str] = None,
    ) -> Iterator[PathOperation]:
        if method == self.ca_method_name:
            yield from super().__call__(
                self.ca_operation_name, paths_iter, spec, base_url=base_url
            )
        else:
            yield from super().__call__(
                method, paths_iter, spec, base_url=base_url
            )


class SimpleServersIterator:
    def __call__(
        self,
        name: str,
        operations_iter: Iterator[PathOperation],
        spec: SchemaPath,
        base_url: Optional[str] = None,
    ) -> Iterator[PathOperationServer]:
        for path, operation, path_result in operations_iter:
            yield PathOperationServer(
                path,
                operation,
                None,
                path_result,
                {},
            )


class TemplateServersIterator:
    def __call__(
        self,
        name: str,
        operations_iter: Iterator[PathOperation],
        spec: SchemaPath,
        base_url: Optional[str] = None,
    ) -> Iterator[PathOperationServer]:
        for path, operation, path_result in operations_iter:
            servers = (
                path.get("servers", None)
                or operation.get("servers", None)
                or spec.get("servers", None)
            )
            if not servers:
                servers = [SchemaPath.from_dict({"url": "/"})]
            for server in servers:
                server_url_pattern = name.rsplit(path_result.resolved, 1)[0]
                server_url = server["url"]
                if not is_absolute(server_url):
                    # relative to absolute url
                    if base_url is not None:
                        server_url = urljoin(base_url, server["url"])
                    # if no base url check only path part
                    else:
                        server_url_pattern = urlparse(server_url_pattern).path
                if server_url.endswith("/"):
                    server_url = server_url[:-1]
                # simple path
                if server_url_pattern == server_url:
                    server_result = TemplateResult(server["url"], {})
                    yield PathOperationServer(
                        path,
                        operation,
                        server,
                        path_result,
                        server_result,
                    )
                # template path
                else:
                    result = parse(server["url"], server_url_pattern)
                    if result:
                        server_result = TemplateResult(
                            server["url"], result.named
                        )
                        yield PathOperationServer(
                            path,
                            operation,
                            server,
                            path_result,
                            server_result,
                        )
                    # servers should'n end with tailing slash
                    # but let's search for this too
                    server_url_pattern += "/"
                    result = parse(server["url"], server_url_pattern)
                    if result:
                        server_result = TemplateResult(
                            server["url"], result.named
                        )
                        yield PathOperationServer(
                            path,
                            operation,
                            server,
                            path_result,
                            server_result,
                        )



"""OpenAPI core contrib starlette requests module"""

from typing import Optional

from starlette.requests import Request

from openapi_core.datatypes import RequestParameters


class StarletteOpenAPIRequest:
    def __init__(self, request: Request, body: Optional[bytes] = None):
        if not isinstance(request, Request):
            raise TypeError(f"'request' argument is not type of {Request}")
        self.request = request

        self.parameters = RequestParameters(
            query=self.request.query_params,
            header=self.request.headers,
            cookie=self.request.cookies,
        )

        self._body = body

    @property
    def host_url(self) -> str:
        return self.request.base_url._url

    @property
    def path(self) -> str:
        return self.request.url.path

    @property
    def method(self) -> str:
        return self.request.method.lower()

    @property
    def body(self) -> Optional[bytes]:
        return self._body

    @property
    def content_type(self) -> str:
        # default value according to RFC 2616
        return (
            self.request.headers.get("Content-Type")
            or "application/octet-stream"
        )



"""OpenAPI core util module"""

from itertools import chain
from typing import Any
from typing import Iterable


def forcebool(val: Any) -> bool:
    if isinstance(val, str):
        val = val.lower()
        if val in ("y", "yes", "t", "true", "on", "1"):
            return True
        elif val in ("n", "no", "f", "false", "off", "0"):
            return False
        else:
            raise ValueError(f"invalid truth value {val!r}")

    return bool(val)


def chainiters(*lists: Iterable[Any]) -> Iterable[Any]:
    iters = map(lambda l: l and iter(l) or [], lists)
    return chain(*iters)



"""OpenAPI core contrib falcon responses module"""

from itertools import tee

from falcon.response import Response
from werkzeug.datastructures import Headers


class FalconOpenAPIResponse:
    def __init__(self, response: Response):
        if not isinstance(response, Response):
            raise TypeError(f"'response' argument is not type of {Response}")
        self.response = response

    @property
    def data(self) -> bytes:
        if self.response.text is None:
            if self.response.stream is None:
                return b""
            resp_iter1, resp_iter2 = tee(self.response.stream)
            self.response.stream = resp_iter1
            content = b"".join(resp_iter2)
            return content
        assert isinstance(self.response.text, str)
        return self.response.text.encode("utf-8")

    @property
    def status_code(self) -> int:
        return int(self.response.status[:3])

    @property
    def content_type(self) -> str:
        content_type = ""
        if self.response.content_type:
            content_type = self.response.content_type
        else:
            content_type = self.response.options.default_media_type
        return content_type

    @property
    def headers(self) -> Headers:
        return Headers(self.response.headers)



from typing import Awaitable
from typing import Callable
from typing import Iterable

from openapi_core.typing import ResponseType
from openapi_core.unmarshalling.request.datatypes import RequestUnmarshalResult

ErrorsHandlerCallable = Callable[[Iterable[Exception]], ResponseType]
ValidRequestHandlerCallable = Callable[[RequestUnmarshalResult], ResponseType]
AsyncValidRequestHandlerCallable = Callable[
    [RequestUnmarshalResult], Awaitable[ResponseType]
]



import logging
from functools import cached_property
from functools import partial
from typing import Any
from typing import Iterator
from typing import Optional

from jsonschema.exceptions import FormatError
from jsonschema.protocols import Validator
from jsonschema_path import SchemaPath

from openapi_core.validation.schemas.datatypes import FormatValidator
from openapi_core.validation.schemas.exceptions import InvalidSchemaValue
from openapi_core.validation.schemas.exceptions import ValidateError

log = logging.getLogger(__name__)


class SchemaValidator:
    def __init__(
        self,
        schema: SchemaPath,
        validator: Validator,
    ):
        self.schema = schema
        self.validator = validator

    def __contains__(self, schema_format: str) -> bool:
        return schema_format in self.validator.format_checker.checkers

    def validate(self, value: Any) -> None:
        errors_iter = self.validator.iter_errors(value)
        errors = tuple(errors_iter)
        if errors:
            schema_type = self.schema.getkey("type", "any")
            raise InvalidSchemaValue(value, schema_type, schema_errors=errors)

    def evolve(self, schema: SchemaPath) -> "SchemaValidator":
        cls = self.__class__

        with schema.resolve() as resolved:
            validator = self.validator.evolve(
                schema=resolved.contents, _resolver=resolved.resolver
            )
            return cls(schema, validator)

    def type_validator(
        self, value: Any, type_override: Optional[str] = None
    ) -> bool:
        callable = self.get_type_validator_callable(
            type_override=type_override
        )
        return callable(value)

    def format_validator(self, value: Any) -> bool:
        try:
            self.format_validator_callable(value)
        except FormatError:
            return False
        else:
            return True

    def get_type_validator_callable(
        self, type_override: Optional[str] = None
    ) -> FormatValidator:
        schema_type = type_override or self.schema.getkey("type")
        if schema_type in self.validator.TYPE_CHECKER._type_checkers:
            return partial(
                self.validator.TYPE_CHECKER.is_type, type=schema_type
            )

        return lambda x: True

    @cached_property
    def format_validator_callable(self) -> FormatValidator:
        schema_format = self.schema.getkey("format")
        if schema_format in self.validator.format_checker.checkers:
            return partial(
                self.validator.format_checker.check, format=schema_format
            )

        return lambda x: True

    def get_primitive_type(self, value: Any) -> Optional[str]:
        schema_types = self.schema.getkey("type")
        if isinstance(schema_types, str):
            return schema_types
        if schema_types is None:
            schema_types = sorted(self.validator.TYPE_CHECKER._type_checkers)
        assert isinstance(schema_types, list)
        for schema_type in schema_types:
            result = self.type_validator(value, type_override=schema_type)
            if not result:
                continue
            result = self.format_validator(value)
            if not result:
                continue
            assert isinstance(schema_type, (str, type(None)))
            return schema_type
        # OpenAPI 3.0: None is not a primitive type so None value will not find any type
        return None

    def iter_valid_schemas(self, value: Any) -> Iterator[SchemaPath]:
        yield self.schema

        one_of_schema = self.get_one_of_schema(value)
        if one_of_schema is not None:
            yield one_of_schema

        yield from self.iter_any_of_schemas(value)
        yield from self.iter_all_of_schemas(value)

    def get_one_of_schema(
        self,
        value: Any,
    ) -> Optional[SchemaPath]:
        if "oneOf" not in self.schema:
            return None

        one_of_schemas = self.schema / "oneOf"
        for subschema in one_of_schemas:
            validator = self.evolve(subschema)
            try:
                validator.validate(value)
            except ValidateError:
                continue
            else:
                return subschema

        log.warning("valid oneOf schema not found")
        return None

    def iter_any_of_schemas(
        self,
        value: Any,
    ) -> Iterator[SchemaPath]:
        if "anyOf" not in self.schema:
            return

        any_of_schemas = self.schema / "anyOf"
        for subschema in any_of_schemas:
            validator = self.evolve(subschema)
            try:
                validator.validate(value)
            except ValidateError:
                continue
            else:
                yield subschema

    def iter_all_of_schemas(
        self,
        value: Any,
    ) -> Iterator[SchemaPath]:
        if "allOf" not in self.schema:
            return

        all_of_schemas = self.schema / "allOf"
        for subschema in all_of_schemas:
            if "type" not in subschema:
                continue
            validator = self.evolve(subschema)
            try:
                validator.validate(value)
            except ValidateError:
                log.warning("invalid allOf schema found")
            else:
                yield subschema



from typing import Any
from typing import Optional

from jsonschema_path import SchemaPath

from openapi_core.protocols import Request
from openapi_core.protocols import Response
from openapi_core.unmarshalling.response.datatypes import (
    ResponseUnmarshalResult,
)
from openapi_core.unmarshalling.response.protocols import ResponseUnmarshaller
from openapi_core.unmarshalling.response.types import ResponseUnmarshallerType


class ResponseUnmarshallingProcessor:
    def __init__(
        self,
        spec: SchemaPath,
        response_unmarshaller_cls: ResponseUnmarshallerType,
        **unmarshaller_kwargs: Any
    ) -> None:
        self.spec = spec
        self.response_unmarshaller_cls = response_unmarshaller_cls
        self.unmarshaller_kwargs = unmarshaller_kwargs

        self._response_unmarshaller_cached: Optional[ResponseUnmarshaller] = (
            None
        )

    @property
    def response_unmarshaller(self) -> ResponseUnmarshaller:
        if self._response_unmarshaller_cached is None:
            self._response_unmarshaller_cached = (
                self.response_unmarshaller_cls(
                    self.spec, **self.unmarshaller_kwargs
                )
            )
        return self._response_unmarshaller_cached

    def process(
        self, request: Request, response: Response
    ) -> ResponseUnmarshalResult:
        return self.response_unmarshaller.unmarshal(request, response)



import warnings
from typing import Any

from jsonschema_path import SchemaPath


class Spec(SchemaPath):
    def __init__(self, *args: Any, **kwargs: Any):
        warnings.warn(
            "Spec is deprecated. Use SchemaPath from jsonschema-path package.",
            DeprecationWarning,
        )
        super().__init__(*args, **kwargs)



from aioitertools.itertools import tee as atee
from starlette.requests import Request
from starlette.responses import Response

from openapi_core.contrib.starlette.requests import StarletteOpenAPIRequest
from openapi_core.contrib.starlette.responses import StarletteOpenAPIResponse
from openapi_core.unmarshalling.processors import AsyncUnmarshallingProcessor
from openapi_core.unmarshalling.typing import ErrorsHandlerCallable


class StarletteIntegration(AsyncUnmarshallingProcessor[Request, Response]):
    request_cls = StarletteOpenAPIRequest
    response_cls = StarletteOpenAPIResponse

    async def get_openapi_request(
        self, request: Request
    ) -> StarletteOpenAPIRequest:
        body = await request.body()
        return self.request_cls(request, body)

    async def get_openapi_response(
        self, response: Response
    ) -> StarletteOpenAPIResponse:
        assert self.response_cls is not None
        data = None
        if hasattr(response, "body_iterator"):
            body_iter1, body_iter2 = atee(response.body_iterator)
            response.body_iterator = body_iter2
            data = b"".join(
                [
                    (
                        chunk.encode(response.charset)
                        if not isinstance(chunk, bytes)
                        else chunk
                    )
                    async for chunk in body_iter1
                ]
            )
        return self.response_cls(response, data=data)

    def should_validate_response(self) -> bool:
        return self.response_cls is not None

    async def handle_response(
        self,
        request: Request,
        response: Response,
        errors_handler: ErrorsHandlerCallable[Response],
    ) -> Response:
        if not self.should_validate_response():
            return response
        return await super().handle_response(request, response, errors_handler)



"""OpenAPI core testing requests module"""

from typing import Any
from typing import Dict
from typing import Optional

from werkzeug.datastructures import Headers
from werkzeug.datastructures import ImmutableMultiDict

from openapi_core.datatypes import RequestParameters


class MockRequest:
    def __init__(
        self,
        host_url: str,
        method: str,
        path: str,
        path_pattern: Optional[str] = None,
        args: Optional[Dict[str, Any]] = None,
        view_args: Optional[Dict[str, Any]] = None,
        headers: Optional[Dict[str, Any]] = None,
        cookies: Optional[Dict[str, Any]] = None,
        data: Optional[bytes] = None,
        content_type: str = "application/json",
    ):
        self.host_url = host_url
        self.method = method.lower()
        self.path = path
        self.path_pattern = path_pattern
        self.args = args
        self.view_args = view_args
        self.headers = headers
        self.cookies = cookies
        self.body = data or b""
        self.content_type = content_type

        self.parameters = RequestParameters(
            path=self.view_args or {},
            query=ImmutableMultiDict(self.args or {}),
            header=Headers(self.headers or {}),
            cookie=ImmutableMultiDict(self.cookies or {}),
        )



from openapi_core.contrib.starlette.requests import StarletteOpenAPIRequest
from openapi_core.contrib.starlette.responses import StarletteOpenAPIResponse

__all__ = [
    "StarletteOpenAPIRequest",
    "StarletteOpenAPIResponse",
]



"""OpenAPI core contrib django requests module"""

import re
from typing import Optional

from django.http.request import HttpRequest
from werkzeug.datastructures import Headers
from werkzeug.datastructures import ImmutableMultiDict

from openapi_core.datatypes import RequestParameters

# https://docs.djangoproject.com/en/stable/topics/http/urls/
#
# Currently unsupported are :
#    - nested arguments, e.g.: ^comments/(?:page-(?P<page_number>\d+)/)?$
#    - unnamed regex groups, e.g.: ^articles/([0-9]{4})/$
#    - multiple named parameters between a single pair of slashes
#     e.g.: <page_slug>-<page_id>/edit/
#
# The regex matches everything, except a "/" until "<". Then only the name
# is exported, after which it matches ">" and everything until a "/".
# A check is made to ensure that "/" is not in an excluded character set such
# as may be found with Django REST Framwork's default value pattern, "[^/.]+".
PATH_PARAMETER_PATTERN = (
    r"(?:[^/]*?)<(?:(?:.*?:))*?(\w+)>(?:(?:[^/]*?\[\^[^/]*/)?[^/]*)"
)


class DjangoOpenAPIRequest:
    path_regex = re.compile(PATH_PARAMETER_PATTERN)

    def __init__(self, request: HttpRequest):
        if not isinstance(request, HttpRequest):
            raise TypeError(f"'request' argument is not type of {HttpRequest}")
        self.request = request

        path = (
            self.request.resolver_match
            and self.request.resolver_match.kwargs
            or {}
        )
        self.parameters = RequestParameters(
            path=path,
            query=ImmutableMultiDict(self.request.GET),
            header=Headers(self.request.headers.items()),
            cookie=ImmutableMultiDict(dict(self.request.COOKIES)),
        )

    @property
    def host_url(self) -> str:
        assert isinstance(self.request._current_scheme_host, str)
        return self.request._current_scheme_host

    @property
    def path(self) -> str:
        assert isinstance(self.request.path, str)
        return self.request.path

    @property
    def path_pattern(self) -> Optional[str]:
        if self.request.resolver_match is None:
            return None

        route = self.path_regex.sub(r"{\1}", self.request.resolver_match.route)
        # Delete start and end marker to allow concatenation.
        if route[:1] == "^":
            route = route[1:]
        if route[-1:] == "$":
            route = route[:-1]
        return "/" + route

    @property
    def method(self) -> str:
        if self.request.method is None:
            return ""
        assert isinstance(self.request.method, str)
        return self.request.method.lower()

    @property
    def body(self) -> bytes:
        assert isinstance(self.request.body, bytes)
        return self.request.body

    @property
    def content_type(self) -> str:
        content_type = self.request.META.get("CONTENT_TYPE", "")
        assert isinstance(content_type, str)
        return content_type



import logging
from typing import Any
from typing import Iterable
from typing import List
from typing import Mapping
from typing import Optional
from typing import Type
from typing import Union

from jsonschema_path import SchemaPath

from openapi_core.extensions.models.factories import ModelPathFactory
from openapi_core.schema.schemas import get_properties
from openapi_core.unmarshalling.schemas.datatypes import FormatUnmarshaller
from openapi_core.unmarshalling.schemas.datatypes import (
    FormatUnmarshallersDict,
)
from openapi_core.validation.schemas.validators import SchemaValidator

log = logging.getLogger(__name__)


class PrimitiveUnmarshaller:
    def __init__(
        self,
        schema: SchemaPath,
        schema_validator: SchemaValidator,
        schema_unmarshaller: "SchemaUnmarshaller",
    ) -> None:
        self.schema = schema
        self.schema_validator = schema_validator
        self.schema_unmarshaller = schema_unmarshaller

    def __call__(self, value: Any) -> Any:
        return value


class ArrayUnmarshaller(PrimitiveUnmarshaller):
    def __call__(self, value: Any) -> Optional[List[Any]]:
        return list(map(self.items_unmarshaller.unmarshal, value))

    @property
    def items_unmarshaller(self) -> "SchemaUnmarshaller":
        # sometimes we don't have any schema i.e. free-form objects
        items_schema = self.schema.get("items", SchemaPath.from_dict({}))
        return self.schema_unmarshaller.evolve(items_schema)


class ObjectUnmarshaller(PrimitiveUnmarshaller):
    def __call__(self, value: Any) -> Any:
        properties = self._unmarshal_properties(value)

        fields: Iterable[str] = properties and properties.keys() or []
        object_class = self.object_class_factory.create(self.schema, fields)

        return object_class(**properties)

    @property
    def object_class_factory(self) -> ModelPathFactory:
        return ModelPathFactory()

    def evolve(self, schema: SchemaPath) -> "ObjectUnmarshaller":
        cls = self.__class__

        return cls(
            schema,
            self.schema_validator.evolve(schema),
            self.schema_unmarshaller,
        )

    def _unmarshal_properties(
        self, value: Any, schema_only: bool = False
    ) -> Any:
        properties = {}

        one_of_schema = self.schema_validator.get_one_of_schema(value)
        if one_of_schema is not None:
            one_of_properties = self.evolve(
                one_of_schema
            )._unmarshal_properties(value, schema_only=True)
            properties.update(one_of_properties)

        any_of_schemas = self.schema_validator.iter_any_of_schemas(value)
        for any_of_schema in any_of_schemas:
            any_of_properties = self.evolve(
                any_of_schema
            )._unmarshal_properties(value, schema_only=True)
            properties.update(any_of_properties)

        all_of_schemas = self.schema_validator.iter_all_of_schemas(value)
        for all_of_schema in all_of_schemas:
            all_of_properties = self.evolve(
                all_of_schema
            )._unmarshal_properties(value, schema_only=True)
            properties.update(all_of_properties)

        for prop_name, prop_schema in get_properties(self.schema).items():
            try:
                prop_value = value[prop_name]
            except KeyError:
                if "default" not in prop_schema:
                    continue
                prop_value = prop_schema["default"]

            properties[prop_name] = self.schema_unmarshaller.evolve(
                prop_schema
            ).unmarshal(prop_value)

        if schema_only:
            return properties

        additional_properties = self.schema.getkey(
            "additionalProperties", True
        )
        if additional_properties is not False:
            # free-form object
            if additional_properties is True:
                additional_prop_schema = SchemaPath.from_dict(
                    {"nullable": True}
                )
            # defined schema
            else:
                additional_prop_schema = self.schema / "additionalProperties"
            additional_prop_unmarshaler = self.schema_unmarshaller.evolve(
                additional_prop_schema
            )
            for prop_name, prop_value in value.items():
                if prop_name in properties:
                    continue
                properties[prop_name] = additional_prop_unmarshaler.unmarshal(
                    prop_value
                )

        return properties


class MultiTypeUnmarshaller(PrimitiveUnmarshaller):
    def __call__(self, value: Any) -> Any:
        primitive_type = self.schema_validator.get_primitive_type(value)
        # OpenAPI 3.0: handle no type for None
        if primitive_type is None:
            return None
        unmarshaller = self.schema_unmarshaller.get_type_unmarshaller(
            primitive_type
        )
        return unmarshaller(value)


class AnyUnmarshaller(MultiTypeUnmarshaller):
    pass


class TypesUnmarshaller:
    unmarshallers: Mapping[str, Type[PrimitiveUnmarshaller]] = {}
    multi: Optional[Type[PrimitiveUnmarshaller]] = None

    def __init__(
        self,
        unmarshallers: Mapping[str, Type[PrimitiveUnmarshaller]],
        default: Type[PrimitiveUnmarshaller],
        multi: Optional[Type[PrimitiveUnmarshaller]] = None,
    ):
        self.unmarshallers = unmarshallers
        self.default = default
        self.multi = multi

    def get_types(self) -> List[str]:
        return list(self.unmarshallers.keys())

    def get_unmarshaller_cls(
        self,
        schema_type: Optional[Union[Iterable[str], str]],
    ) -> Type["PrimitiveUnmarshaller"]:
        if schema_type is None:
            return self.default
        if isinstance(schema_type, Iterable) and not isinstance(
            schema_type, str
        ):
            if self.multi is None:
                raise TypeError("Unmarshaller does not accept multiple types")
            return self.multi

        return self.unmarshallers[schema_type]


class FormatsUnmarshaller:
    def __init__(
        self,
        format_unmarshallers: Optional[FormatUnmarshallersDict] = None,
        extra_format_unmarshallers: Optional[FormatUnmarshallersDict] = None,
    ):
        if format_unmarshallers is None:
            format_unmarshallers = {}
        self.format_unmarshallers = format_unmarshallers
        if extra_format_unmarshallers is None:
            extra_format_unmarshallers = {}
        self.extra_format_unmarshallers = extra_format_unmarshallers

    def unmarshal(self, schema_format: str, value: Any) -> Any:
        format_unmarshaller = self.get_unmarshaller(schema_format)
        if format_unmarshaller is None:
            return value
        try:
            return format_unmarshaller(value)
        except (AttributeError, ValueError, TypeError):
            return value

    def get_unmarshaller(
        self, schema_format: str
    ) -> Optional[FormatUnmarshaller]:
        if schema_format in self.extra_format_unmarshallers:
            return self.extra_format_unmarshallers[schema_format]
        if schema_format in self.format_unmarshallers:
            return self.format_unmarshallers[schema_format]

        return None

    def __contains__(self, schema_format: str) -> bool:
        format_unmarshallers_dicts: List[Mapping[str, Any]] = [
            self.extra_format_unmarshallers,
            self.format_unmarshallers,
        ]
        for content in format_unmarshallers_dicts:
            if schema_format in content:
                return True
        return False


class SchemaUnmarshaller:
    def __init__(
        self,
        schema: SchemaPath,
        schema_validator: SchemaValidator,
        types_unmarshaller: TypesUnmarshaller,
        formats_unmarshaller: FormatsUnmarshaller,
    ):
        self.schema = schema
        self.schema_validator = schema_validator

        self.types_unmarshaller = types_unmarshaller
        self.formats_unmarshaller = formats_unmarshaller

    def unmarshal(self, value: Any) -> Any:
        self.schema_validator.validate(value)

        # skip unmarshalling for nullable in OpenAPI 3.0
        if value is None and self.schema.getkey("nullable", False):
            return value

        schema_type = self.schema.getkey("type")
        type_unmarshaller = self.get_type_unmarshaller(schema_type)
        typed = type_unmarshaller(value)
        # skip finding format for None
        if typed is None:
            return None
        schema_format = self.find_format(value)
        if schema_format is None:
            return typed
        # ignore incompatible formats
        if not (
            isinstance(value, str)
            or
            # Workaround allows bytes for binary and byte formats
            (isinstance(value, bytes) and schema_format in ["binary", "byte"])
        ):
            return typed

        format_unmarshaller = self.get_format_unmarshaller(schema_format)
        if format_unmarshaller is None:
            return typed
        try:
            return format_unmarshaller(typed)
        except (AttributeError, ValueError, TypeError):
            return typed

    def get_type_unmarshaller(
        self,
        schema_type: Optional[Union[Iterable[str], str]],
    ) -> PrimitiveUnmarshaller:
        klass = self.types_unmarshaller.get_unmarshaller_cls(schema_type)
        return klass(
            self.schema,
            self.schema_validator,
            self,
        )

    def get_format_unmarshaller(
        self,
        schema_format: str,
    ) -> Optional[FormatUnmarshaller]:
        return self.formats_unmarshaller.get_unmarshaller(schema_format)

    def evolve(self, schema: SchemaPath) -> "SchemaUnmarshaller":
        cls = self.__class__

        return cls(
            schema,
            self.schema_validator.evolve(schema),
            self.types_unmarshaller,
            self.formats_unmarshaller,
        )

    def find_format(self, value: Any) -> Optional[str]:
        for schema in self.schema_validator.iter_valid_schemas(value):
            schema_validator = self.schema_validator.evolve(schema)
            primitive_type = schema_validator.get_primitive_type(value)
            if primitive_type != "string":
                continue
            if "format" in schema:
                return str(schema.getkey("format"))
        return None



from django.http.request import HttpRequest
from django.http.response import HttpResponse

from openapi_core.contrib.django.requests import DjangoOpenAPIRequest
from openapi_core.contrib.django.responses import DjangoOpenAPIResponse
from openapi_core.unmarshalling.processors import UnmarshallingProcessor
from openapi_core.unmarshalling.typing import ErrorsHandlerCallable


class DjangoIntegration(UnmarshallingProcessor[HttpRequest, HttpResponse]):
    request_cls = DjangoOpenAPIRequest
    response_cls = DjangoOpenAPIResponse

    def get_openapi_request(
        self, request: HttpRequest
    ) -> DjangoOpenAPIRequest:
        return self.request_cls(request)

    def get_openapi_response(
        self, response: HttpResponse
    ) -> DjangoOpenAPIResponse:
        assert self.response_cls is not None
        return self.response_cls(response)

    def should_validate_response(self) -> bool:
        return self.response_cls is not None

    def handle_response(
        self,
        request: HttpRequest,
        response: HttpResponse,
        errors_handler: ErrorsHandlerCallable[HttpResponse],
    ) -> HttpResponse:
        if not self.should_validate_response():
            return response
        return super().handle_response(request, response, errors_handler)



"""OpenAPI core contrib flask handlers module"""

from typing import Any
from typing import Callable
from typing import Dict
from typing import Iterable
from typing import Type

from flask.globals import current_app
from flask.helpers import make_response
from flask.json import dumps
from flask.wrappers import Request
from flask.wrappers import Response

from openapi_core.templating.media_types.exceptions import MediaTypeNotFound
from openapi_core.templating.paths.exceptions import OperationNotFound
from openapi_core.templating.paths.exceptions import PathNotFound
from openapi_core.templating.paths.exceptions import ServerNotFound
from openapi_core.templating.security.exceptions import SecurityNotFound
from openapi_core.unmarshalling.request.datatypes import RequestUnmarshalResult


class FlaskOpenAPIErrorsHandler:
    OPENAPI_ERROR_STATUS: Dict[Type[BaseException], int] = {
        ServerNotFound: 400,
        SecurityNotFound: 403,
        OperationNotFound: 405,
        PathNotFound: 404,
        MediaTypeNotFound: 415,
    }

    def __call__(self, errors: Iterable[Exception]) -> Response:
        data_errors = [self.format_openapi_error(err) for err in errors]
        data = {
            "errors": data_errors,
        }
        data_error_max = max(data_errors, key=self.get_error_status)
        status = data_error_max["status"]
        return current_app.response_class(
            dumps(data), status=status, mimetype="application/json"
        )

    @classmethod
    def format_openapi_error(cls, error: BaseException) -> Dict[str, Any]:
        if error.__cause__ is not None:
            error = error.__cause__
        return {
            "title": str(error),
            "status": cls.OPENAPI_ERROR_STATUS.get(error.__class__, 400),
            "class": str(type(error)),
        }

    @classmethod
    def get_error_status(cls, error: Dict[str, Any]) -> int:
        return int(error["status"])


class FlaskOpenAPIValidRequestHandler:
    def __init__(
        self,
        req: Request,
        view: Callable[[Any], Response],
        *view_args: Any,
        **view_kwargs: Any,
    ):
        self.req = req
        self.view = view
        self.view_args = view_args
        self.view_kwargs = view_kwargs

    def __call__(
        self, request_unmarshal_result: RequestUnmarshalResult
    ) -> Response:
        self.req.openapi = request_unmarshal_result  # type: ignore
        rv = self.view(*self.view_args, **self.view_kwargs)
        return make_response(rv)



from typing import Protocol
from typing import runtime_checkable

from requests.cookies import RequestsCookieJar


@runtime_checkable
class SupportsCookieJar(Protocol):
    _cookies: RequestsCookieJar



"""OpenAPI core contrib django providers module"""

import warnings
from typing import cast

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured

from openapi_core import OpenAPI


def get_default_openapi_instance() -> OpenAPI:
    """
    Retrieves or initializes the OpenAPI instance based on Django settings
    (either OPENAPI or OPENAPI_SPEC).
    This function ensures the spec is only loaded once.
    """
    if hasattr(settings, "OPENAPI"):
        # Recommended (newer) approach
        return cast(OpenAPI, settings.OPENAPI)
    elif hasattr(settings, "OPENAPI_SPEC"):
        # Backward compatibility
        warnings.warn(
            "OPENAPI_SPEC is deprecated. Use OPENAPI in your settings instead.",
            DeprecationWarning,
        )
        return OpenAPI(settings.OPENAPI_SPEC)
    else:
        raise ImproperlyConfigured(
            "Neither OPENAPI nor OPENAPI_SPEC is defined in Django settings."
        )



from typing import Optional

from jsonschema_path import SchemaPath

from openapi_core.casting.schemas.casters import SchemaCaster
from openapi_core.casting.schemas.casters import TypesCaster
from openapi_core.validation.schemas.datatypes import FormatValidatorsDict
from openapi_core.validation.schemas.factories import SchemaValidatorsFactory


class SchemaCastersFactory:
    def __init__(
        self,
        schema_validators_factory: SchemaValidatorsFactory,
        types_caster: TypesCaster,
    ):
        self.schema_validators_factory = schema_validators_factory
        self.types_caster = types_caster

    def create(
        self,
        schema: SchemaPath,
        format_validators: Optional[FormatValidatorsDict] = None,
        extra_format_validators: Optional[FormatValidatorsDict] = None,
    ) -> SchemaCaster:
        schema_validator = self.schema_validators_factory.create(
            schema,
            format_validators=format_validators,
            extra_format_validators=extra_format_validators,
        )

        return SchemaCaster(schema, schema_validator, self.types_caster)



from dataclasses import dataclass
from typing import Iterable

from jsonschema_path import SchemaPath

from openapi_core.datatypes import Parameters
from openapi_core.exceptions import OpenAPIError
from openapi_core.validation.exceptions import ValidationError
from openapi_core.validation.schemas.exceptions import ValidateError


@dataclass
class ParametersError(Exception):
    parameters: Parameters
    errors: Iterable[OpenAPIError]


class RequestValidationError(ValidationError):
    """Request validation error"""


class RequestBodyValidationError(RequestValidationError):
    def __str__(self) -> str:
        return "Request body validation error"


class InvalidRequestBody(RequestBodyValidationError, ValidateError):
    """Invalid request body"""


class MissingRequestBodyError(RequestBodyValidationError):
    """Missing request body error"""


class MissingRequestBody(MissingRequestBodyError):
    def __str__(self) -> str:
        return "Missing request body"


class MissingRequiredRequestBody(MissingRequestBodyError):
    def __str__(self) -> str:
        return "Missing required request body"


@dataclass
class ParameterValidationError(RequestValidationError):
    name: str
    location: str

    @classmethod
    def from_spec(cls, spec: SchemaPath) -> "ParameterValidationError":
        return cls(spec["name"], spec["in"])

    def __str__(self) -> str:
        return f"{self.location.title()} parameter error: {self.name}"


class InvalidParameter(ParameterValidationError, ValidateError):
    def __str__(self) -> str:
        return f"Invalid {self.location} parameter: {self.name}"


class MissingParameterError(ParameterValidationError):
    """Missing parameter error"""


class MissingParameter(MissingParameterError):
    def __str__(self) -> str:
        return f"Missing {self.location} parameter: {self.name}"


class MissingRequiredParameter(MissingParameterError):
    def __str__(self) -> str:
        return f"Missing required {self.location} parameter: {self.name}"


class SecurityValidationError(RequestValidationError):
    pass


class InvalidSecurity(SecurityValidationError, ValidateError):
    """Invalid security"""



from flask.wrappers import Request
from flask.wrappers import Response

from openapi_core.contrib.flask.requests import FlaskOpenAPIRequest
from openapi_core.contrib.flask.responses import FlaskOpenAPIResponse
from openapi_core.unmarshalling.processors import UnmarshallingProcessor
from openapi_core.unmarshalling.typing import ErrorsHandlerCallable


class FlaskIntegration(UnmarshallingProcessor[Request, Response]):
    request_cls = FlaskOpenAPIRequest
    response_cls = FlaskOpenAPIResponse

    def get_openapi_request(self, request: Request) -> FlaskOpenAPIRequest:
        return self.request_cls(request)

    def get_openapi_response(self, response: Response) -> FlaskOpenAPIResponse:
        assert self.response_cls is not None
        return self.response_cls(response)

    def should_validate_response(self) -> bool:
        return self.response_cls is not None

    def handle_response(
        self,
        request: Request,
        response: Response,
        errors_handler: ErrorsHandlerCallable[Response],
    ) -> Response:
        if not self.should_validate_response():
            return response
        return super().handle_response(request, response, errors_handler)



"""OpenAPI core contrib django middlewares module"""

from typing import Callable

from django.conf import settings
from django.http.request import HttpRequest
from django.http.response import HttpResponse

from openapi_core.contrib.django.handlers import DjangoOpenAPIErrorsHandler
from openapi_core.contrib.django.handlers import (
    DjangoOpenAPIValidRequestHandler,
)
from openapi_core.contrib.django.integrations import DjangoIntegration
from openapi_core.contrib.django.providers import get_default_openapi_instance


class DjangoOpenAPIMiddleware(DjangoIntegration):
    valid_request_handler_cls = DjangoOpenAPIValidRequestHandler
    errors_handler = DjangoOpenAPIErrorsHandler()

    def __init__(self, get_response: Callable[[HttpRequest], HttpResponse]):
        self.get_response = get_response

        if hasattr(settings, "OPENAPI_RESPONSE_CLS"):
            self.response_cls = settings.OPENAPI_RESPONSE_CLS

        openapi = get_default_openapi_instance()

        super().__init__(openapi)

    def __call__(self, request: HttpRequest) -> HttpResponse:
        valid_request_handler = self.valid_request_handler_cls(
            request, self.get_response
        )
        response = self.handle_request(
            request, valid_request_handler, self.errors_handler
        )

        return self.handle_response(request, response, self.errors_handler)



from openapi_core.templating.paths.finders import APICallPathFinder
from openapi_core.templating.paths.finders import WebhookPathFinder

__all__ = [
    "APICallPathFinder",
    "WebhookPathFinder",
]



from typing import Any
from typing import Dict
from typing import Type

from jsonschema_path import SchemaPath

from openapi_core.security.providers import ApiKeyProvider
from openapi_core.security.providers import BaseProvider
from openapi_core.security.providers import HttpProvider
from openapi_core.security.providers import UnsupportedProvider


class SecurityProviderFactory:
    PROVIDERS: Dict[str, Type[BaseProvider]] = {
        "apiKey": ApiKeyProvider,
        "http": HttpProvider,
        "oauth2": UnsupportedProvider,
        "openIdConnect": UnsupportedProvider,
    }

    def create(self, scheme: SchemaPath) -> Any:
        scheme_type = scheme["type"]
        provider_class = self.PROVIDERS[scheme_type]
        return provider_class(scheme)



"""OpenAPI core unmarshalling processors module"""

from typing import Generic

from openapi_core.app import OpenAPI
from openapi_core.protocols import Request
from openapi_core.protocols import Response
from openapi_core.typing import RequestType
from openapi_core.typing import ResponseType
from openapi_core.unmarshalling.request.datatypes import RequestUnmarshalResult
from openapi_core.unmarshalling.response.datatypes import (
    ResponseUnmarshalResult,
)
from openapi_core.validation.integrations import ValidationIntegration


class UnmarshallingIntegration(
    ValidationIntegration[RequestType, ResponseType]
):
    def unmarshal_request(
        self, request: RequestType
    ) -> RequestUnmarshalResult:
        openapi_request = self.get_openapi_request(request)
        return self.openapi.unmarshal_request(
            openapi_request,
        )

    def unmarshal_response(
        self,
        request: RequestType,
        response: ResponseType,
    ) -> ResponseUnmarshalResult:
        openapi_request = self.get_openapi_request(request)
        openapi_response = self.get_openapi_response(response)
        return self.openapi.unmarshal_response(
            openapi_request, openapi_response
        )


class AsyncUnmarshallingIntegration(Generic[RequestType, ResponseType]):
    def __init__(
        self,
        openapi: OpenAPI,
    ):
        self.openapi = openapi

    async def get_openapi_request(self, request: RequestType) -> Request:
        raise NotImplementedError

    async def get_openapi_response(self, response: ResponseType) -> Response:
        raise NotImplementedError

    async def unmarshal_request(
        self,
        request: RequestType,
    ) -> RequestUnmarshalResult:
        openapi_request = await self.get_openapi_request(request)
        return self.openapi.unmarshal_request(openapi_request)

    async def unmarshal_response(
        self,
        request: RequestType,
        response: ResponseType,
    ) -> ResponseUnmarshalResult:
        openapi_request = await self.get_openapi_request(request)
        openapi_response = await self.get_openapi_response(response)
        return self.openapi.unmarshal_response(
            openapi_request, openapi_response
        )



"""OpenAPI core protocols"""

from typing import Any
from typing import Mapping
from typing import Optional
from typing import Protocol
from typing import runtime_checkable

from openapi_core.datatypes import RequestParameters


@runtime_checkable
class BaseRequest(Protocol):
    parameters: RequestParameters

    @property
    def method(self) -> str:
        """The request method, as lowercase string."""

    @property
    def body(self) -> Optional[bytes]:
        """The request body, as bytes (None if not provided)."""

    @property
    def content_type(self) -> str:
        """The content type with parameters (e.g., charset, boundary, etc.) and always lowercase."""


@runtime_checkable
class Request(BaseRequest, Protocol):
    """Request protocol.

    Attributes:
        host_url: Url with scheme and host.
            For example: https://localhost:8000
        path: Request path.
        full_url_pattern: The matched url with scheme, host and path pattern.
            For example: https://localhost:8000/api/v1/pets
                         https://localhost:8000/api/v1/pets/{pet_id}
        method: The request method, as lowercase string.
        parameters: A RequestParameters object. Needs to support path attribute setter
            to write resolved path parameters.
        content_type: The content type with parameters (e.g., charset, boundary, etc.)
            and always lowercase.
        body: The request body, as bytes (None if not provided).
    """

    @property
    def host_url(self) -> str:
        """Url with scheme and host. For example: https://localhost:8000"""

    @property
    def path(self) -> str:
        """Request path."""


@runtime_checkable
class WebhookRequest(BaseRequest, Protocol):
    """Webhook request protocol.

    Attributes:
        name: Webhook name.
        method: The request method, as lowercase string.
        parameters: A RequestParameters object. Needs to support path attribute setter
            to write resolved path parameters.
        content_type: The content type with parameters (e.g., charset, boundary, etc.)
            and always lowercase.
        body: The request body, as bytes (None if not provided).
    """

    @property
    def name(self) -> str:
        """Webhook name."""


@runtime_checkable
class SupportsPathPattern(Protocol):
    """Supports path_pattern protocol.

    You also need to provide path variables in RequestParameters.

    Attributes:
        path_pattern: The matched path pattern.
            For example: /api/v1/pets/{pet_id}
    """

    @property
    def path_pattern(self) -> str:
        """The matched path pattern. For example: /api/v1/pets/{pet_id}"""


@runtime_checkable
class Response(Protocol):
    """Response protocol.

    Attributes:
        status_code: The status code as integer.
        headers: Response headers as Headers.
        content_type: The content type with parameters and always lowercase.
        data: The response body, as bytes (None if not provided).
    """

    @property
    def status_code(self) -> int:
        """The status code as integer."""

    @property
    def content_type(self) -> str:
        """The content type with parameters and always lowercase."""

    @property
    def headers(self) -> Mapping[str, Any]:
        """Response headers as Headers."""

    @property
    def data(self) -> Optional[bytes]:
        """The response body, as bytes (None if not provided)."""



from openapi_core.contrib.requests.requests import RequestsOpenAPIRequest
from openapi_core.contrib.requests.requests import (
    RequestsOpenAPIWebhookRequest,
)
from openapi_core.contrib.requests.responses import RequestsOpenAPIResponse

__all__ = [
    "RequestsOpenAPIRequest",
    "RequestsOpenAPIResponse",
    "RequestsOpenAPIWebhookRequest",
]



from collections import defaultdict

from openapi_core.deserializing.media_types.datatypes import (
    MediaTypeDeserializersDict,
)
from openapi_core.deserializing.media_types.factories import (
    MediaTypeDeserializersFactory,
)
from openapi_core.deserializing.media_types.util import binary_loads
from openapi_core.deserializing.media_types.util import data_form_loads
from openapi_core.deserializing.media_types.util import json_loads
from openapi_core.deserializing.media_types.util import plain_loads
from openapi_core.deserializing.media_types.util import urlencoded_form_loads
from openapi_core.deserializing.media_types.util import xml_loads
from openapi_core.deserializing.styles import style_deserializers_factory

__all__ = ["media_type_deserializers_factory"]

media_type_deserializers: MediaTypeDeserializersDict = defaultdict(
    lambda: binary_loads,
    **{
        "text/html": plain_loads,
        "text/plain": plain_loads,
        "application/octet-stream": binary_loads,
        "application/json": json_loads,
        "application/vnd.api+json": json_loads,
        "application/xml": xml_loads,
        "application/xhtml+xml": xml_loads,
        "application/x-www-form-urlencoded": urlencoded_form_loads,
        "multipart/form-data": data_form_loads,
    }
)

media_type_deserializers_factory = MediaTypeDeserializersFactory(
    style_deserializers_factory,
    media_type_deserializers=media_type_deserializers,
)



"""OpenAPI core contrib django responses module"""

from itertools import tee

from django.http.response import HttpResponse
from django.http.response import StreamingHttpResponse
from werkzeug.datastructures import Headers


class DjangoOpenAPIResponse:
    def __init__(self, response: HttpResponse):
        if not isinstance(response, (HttpResponse, StreamingHttpResponse)):
            raise TypeError(
                f"'response' argument is not type of {HttpResponse} or {StreamingHttpResponse}"
            )
        self.response = response

    @property
    def data(self) -> bytes:
        if isinstance(self.response, StreamingHttpResponse):
            resp_iter1, resp_iter2 = tee(self.response._iterator)
            self.response.streaming_content = resp_iter1
            content = b"".join(map(self.response.make_bytes, resp_iter2))
            return content
        assert isinstance(self.response.content, bytes)
        return self.response.content

    @property
    def status_code(self) -> int:
        assert isinstance(self.response.status_code, int)
        return self.response.status_code

    @property
    def headers(self) -> Headers:
        return Headers(self.response.headers.items())

    @property
    def content_type(self) -> str:
        content_type = self.response.get("Content-Type", "")
        assert isinstance(content_type, str)
        return content_type



from typing import Optional

from jsonschema_path import SchemaPath
from openapi_spec_validator.validation.types import SpecValidatorType

from openapi_core.casting.schemas.factories import SchemaCastersFactory
from openapi_core.deserializing.media_types import (
    media_type_deserializers_factory,
)
from openapi_core.deserializing.media_types.datatypes import (
    MediaTypeDeserializersDict,
)
from openapi_core.deserializing.media_types.factories import (
    MediaTypeDeserializersFactory,
)
from openapi_core.deserializing.styles import style_deserializers_factory
from openapi_core.deserializing.styles.factories import (
    StyleDeserializersFactory,
)
from openapi_core.protocols import BaseRequest
from openapi_core.protocols import Request
from openapi_core.protocols import WebhookRequest
from openapi_core.security import security_provider_factory
from openapi_core.security.factories import SecurityProviderFactory
from openapi_core.templating.paths.exceptions import PathError
from openapi_core.templating.paths.types import PathFinderType
from openapi_core.unmarshalling.request.datatypes import RequestUnmarshalResult
from openapi_core.unmarshalling.schemas import (
    oas30_write_schema_unmarshallers_factory,
)
from openapi_core.unmarshalling.schemas import (
    oas31_schema_unmarshallers_factory,
)
from openapi_core.unmarshalling.schemas.datatypes import (
    FormatUnmarshallersDict,
)
from openapi_core.unmarshalling.schemas.factories import (
    SchemaUnmarshallersFactory,
)
from openapi_core.unmarshalling.unmarshallers import BaseUnmarshaller
from openapi_core.util import chainiters
from openapi_core.validation.request.exceptions import MissingRequestBody
from openapi_core.validation.request.exceptions import ParametersError
from openapi_core.validation.request.exceptions import (
    RequestBodyValidationError,
)
from openapi_core.validation.request.exceptions import SecurityValidationError
from openapi_core.validation.request.validators import APICallRequestValidator
from openapi_core.validation.request.validators import BaseRequestValidator
from openapi_core.validation.request.validators import V30RequestBodyValidator
from openapi_core.validation.request.validators import (
    V30RequestParametersValidator,
)
from openapi_core.validation.request.validators import (
    V30RequestSecurityValidator,
)
from openapi_core.validation.request.validators import V30RequestValidator
from openapi_core.validation.request.validators import V31RequestBodyValidator
from openapi_core.validation.request.validators import (
    V31RequestParametersValidator,
)
from openapi_core.validation.request.validators import (
    V31RequestSecurityValidator,
)
from openapi_core.validation.request.validators import V31RequestValidator
from openapi_core.validation.request.validators import (
    V31WebhookRequestBodyValidator,
)
from openapi_core.validation.request.validators import (
    V31WebhookRequestParametersValidator,
)
from openapi_core.validation.request.validators import (
    V31WebhookRequestSecurityValidator,
)
from openapi_core.validation.request.validators import (
    V31WebhookRequestValidator,
)
from openapi_core.validation.request.validators import WebhookRequestValidator
from openapi_core.validation.schemas.datatypes import FormatValidatorsDict
from openapi_core.validation.schemas.factories import SchemaValidatorsFactory


class BaseRequestUnmarshaller(BaseRequestValidator, BaseUnmarshaller):
    def __init__(
        self,
        spec: SchemaPath,
        base_url: Optional[str] = None,
        style_deserializers_factory: StyleDeserializersFactory = style_deserializers_factory,
        media_type_deserializers_factory: MediaTypeDeserializersFactory = media_type_deserializers_factory,
        schema_casters_factory: Optional[SchemaCastersFactory] = None,
        schema_validators_factory: Optional[SchemaValidatorsFactory] = None,
        path_finder_cls: Optional[PathFinderType] = None,
        spec_validator_cls: Optional[SpecValidatorType] = None,
        format_validators: Optional[FormatValidatorsDict] = None,
        extra_format_validators: Optional[FormatValidatorsDict] = None,
        extra_media_type_deserializers: Optional[
            MediaTypeDeserializersDict
        ] = None,
        security_provider_factory: SecurityProviderFactory = security_provider_factory,
        schema_unmarshallers_factory: Optional[
            SchemaUnmarshallersFactory
        ] = None,
        format_unmarshallers: Optional[FormatUnmarshallersDict] = None,
        extra_format_unmarshallers: Optional[FormatUnmarshallersDict] = None,
    ):
        BaseUnmarshaller.__init__(
            self,
            spec,
            base_url=base_url,
            style_deserializers_factory=style_deserializers_factory,
            media_type_deserializers_factory=media_type_deserializers_factory,
            schema_casters_factory=schema_casters_factory,
            schema_validators_factory=schema_validators_factory,
            path_finder_cls=path_finder_cls,
            spec_validator_cls=spec_validator_cls,
            format_validators=format_validators,
            extra_format_validators=extra_format_validators,
            extra_media_type_deserializers=extra_media_type_deserializers,
            schema_unmarshallers_factory=schema_unmarshallers_factory,
            format_unmarshallers=format_unmarshallers,
            extra_format_unmarshallers=extra_format_unmarshallers,
        )
        BaseRequestValidator.__init__(
            self,
            spec,
            base_url=base_url,
            style_deserializers_factory=style_deserializers_factory,
            media_type_deserializers_factory=media_type_deserializers_factory,
            schema_casters_factory=schema_casters_factory,
            schema_validators_factory=schema_validators_factory,
            path_finder_cls=path_finder_cls,
            spec_validator_cls=spec_validator_cls,
            format_validators=format_validators,
            extra_format_validators=extra_format_validators,
            extra_media_type_deserializers=extra_media_type_deserializers,
            security_provider_factory=security_provider_factory,
        )

    def _unmarshal(
        self, request: BaseRequest, operation: SchemaPath, path: SchemaPath
    ) -> RequestUnmarshalResult:
        try:
            security = self._get_security(request.parameters, operation)
        except SecurityValidationError as exc:
            return RequestUnmarshalResult(errors=[exc])

        try:
            params = self._get_parameters(request.parameters, operation, path)
        except ParametersError as exc:
            params = exc.parameters
            params_errors = exc.errors
        else:
            params_errors = []

        try:
            body = self._get_body(
                request.body, request.content_type, operation
            )
        except MissingRequestBody:
            body = None
            body_errors = []
        except RequestBodyValidationError as exc:
            body = None
            body_errors = [exc]
        else:
            body_errors = []

        errors = list(chainiters(params_errors, body_errors))
        return RequestUnmarshalResult(
            errors=errors,
            body=body,
            parameters=params,
            security=security,
        )

    def _unmarshal_body(
        self, request: BaseRequest, operation: SchemaPath, path: SchemaPath
    ) -> RequestUnmarshalResult:
        try:
            body = self._get_body(
                request.body, request.content_type, operation
            )
        except MissingRequestBody:
            body = None
            errors = []
        except RequestBodyValidationError as exc:
            body = None
            errors = [exc]
        else:
            errors = []

        return RequestUnmarshalResult(
            errors=errors,
            body=body,
        )

    def _unmarshal_parameters(
        self, request: BaseRequest, operation: SchemaPath, path: SchemaPath
    ) -> RequestUnmarshalResult:
        try:
            params = self._get_parameters(request.parameters, path, operation)
        except ParametersError as exc:
            params = exc.parameters
            params_errors = exc.errors
        else:
            params_errors = []

        return RequestUnmarshalResult(
            errors=params_errors,
            parameters=params,
        )

    def _unmarshal_security(
        self, request: BaseRequest, operation: SchemaPath, path: SchemaPath
    ) -> RequestUnmarshalResult:
        try:
            security = self._get_security(request.parameters, operation)
        except SecurityValidationError as exc:
            return RequestUnmarshalResult(errors=[exc])

        return RequestUnmarshalResult(
            errors=[],
            security=security,
        )


class BaseAPICallRequestUnmarshaller(BaseRequestUnmarshaller):
    pass


class BaseWebhookRequestUnmarshaller(BaseRequestUnmarshaller):
    pass


class APICallRequestUnmarshaller(
    APICallRequestValidator, BaseAPICallRequestUnmarshaller
):
    def unmarshal(self, request: Request) -> RequestUnmarshalResult:
        try:
            path, operation, _, path_result, _ = self._find_path(request)
        # don't process if operation errors
        except PathError as exc:
            return RequestUnmarshalResult(errors=[exc])

        request.parameters.path = (
            request.parameters.path or path_result.variables
        )

        return self._unmarshal(request, operation, path)


class APICallRequestBodyUnmarshaller(
    APICallRequestValidator, BaseAPICallRequestUnmarshaller
):
    def unmarshal(self, request: Request) -> RequestUnmarshalResult:
        try:
            path, operation, _, path_result, _ = self._find_path(request)
        # don't process if operation errors
        except PathError as exc:
            return RequestUnmarshalResult(errors=[exc])

        request.parameters.path = (
            request.parameters.path or path_result.variables
        )

        return self._unmarshal_body(request, operation, path)


class APICallRequestParametersUnmarshaller(
    APICallRequestValidator, BaseAPICallRequestUnmarshaller
):
    def unmarshal(self, request: Request) -> RequestUnmarshalResult:
        try:
            path, operation, _, path_result, _ = self._find_path(request)
        # don't process if operation errors
        except PathError as exc:
            return RequestUnmarshalResult(errors=[exc])

        request.parameters.path = (
            request.parameters.path or path_result.variables
        )

        return self._unmarshal_parameters(request, operation, path)


class APICallRequestSecurityUnmarshaller(
    APICallRequestValidator, BaseAPICallRequestUnmarshaller
):
    def unmarshal(self, request: Request) -> RequestUnmarshalResult:
        try:
            path, operation, _, path_result, _ = self._find_path(request)
        # don't process if operation errors
        except PathError as exc:
            return RequestUnmarshalResult(errors=[exc])

        request.parameters.path = (
            request.parameters.path or path_result.variables
        )

        return self._unmarshal_security(request, operation, path)


class WebhookRequestUnmarshaller(
    WebhookRequestValidator, BaseWebhookRequestUnmarshaller
):
    def unmarshal(self, request: WebhookRequest) -> RequestUnmarshalResult:
        try:
            path, operation, _, path_result, _ = self._find_path(request)
        # don't process if operation errors
        except PathError as exc:
            return RequestUnmarshalResult(errors=[exc])

        request.parameters.path = (
            request.parameters.path or path_result.variables
        )

        return self._unmarshal(request, operation, path)


class WebhookRequestBodyUnmarshaller(
    WebhookRequestValidator, BaseWebhookRequestUnmarshaller
):
    def unmarshal(self, request: WebhookRequest) -> RequestUnmarshalResult:
        try:
            path, operation, _, path_result, _ = self._find_path(request)
        # don't process if operation errors
        except PathError as exc:
            return RequestUnmarshalResult(errors=[exc])

        request.parameters.path = (
            request.parameters.path or path_result.variables
        )

        return self._unmarshal_body(request, operation, path)


class WebhookRequestParametersUnmarshaller(
    WebhookRequestValidator, BaseWebhookRequestUnmarshaller
):
    def unmarshal(self, request: WebhookRequest) -> RequestUnmarshalResult:
        try:
            path, operation, _, path_result, _ = self._find_path(request)
        # don't process if operation errors
        except PathError as exc:
            return RequestUnmarshalResult(errors=[exc])

        request.parameters.path = (
            request.parameters.path or path_result.variables
        )

        return self._unmarshal_parameters(request, operation, path)


class WebhookRequestSecuritysUnmarshaller(
    WebhookRequestValidator, BaseWebhookRequestUnmarshaller
):
    def unmarshal(self, request: WebhookRequest) -> RequestUnmarshalResult:
        try:
            path, operation, _, path_result, _ = self._find_path(request)
        # don't process if operation errors
        except PathError as exc:
            return RequestUnmarshalResult(errors=[exc])

        request.parameters.path = (
            request.parameters.path or path_result.variables
        )

        return self._unmarshal_security(request, operation, path)


class V30RequestBodyUnmarshaller(
    V30RequestBodyValidator, APICallRequestBodyUnmarshaller
):
    schema_unmarshallers_factory = oas30_write_schema_unmarshallers_factory


class V30RequestParametersUnmarshaller(
    V30RequestParametersValidator, APICallRequestParametersUnmarshaller
):
    schema_unmarshallers_factory = oas30_write_schema_unmarshallers_factory


class V30RequestSecurityUnmarshaller(
    V30RequestSecurityValidator, APICallRequestSecurityUnmarshaller
):
    schema_unmarshallers_factory = oas30_write_schema_unmarshallers_factory


class V30RequestUnmarshaller(V30RequestValidator, APICallRequestUnmarshaller):
    schema_unmarshallers_factory = oas30_write_schema_unmarshallers_factory


class V31RequestBodyUnmarshaller(
    V31RequestBodyValidator, APICallRequestBodyUnmarshaller
):
    schema_unmarshallers_factory = oas31_schema_unmarshallers_factory


class V31RequestParametersUnmarshaller(
    V31RequestParametersValidator, APICallRequestParametersUnmarshaller
):
    schema_unmarshallers_factory = oas31_schema_unmarshallers_factory


class V31RequestSecurityUnmarshaller(
    V31RequestSecurityValidator, APICallRequestSecurityUnmarshaller
):
    schema_unmarshallers_factory = oas31_schema_unmarshallers_factory


class V31RequestUnmarshaller(V31RequestValidator, APICallRequestUnmarshaller):
    schema_unmarshallers_factory = oas31_schema_unmarshallers_factory


class V31WebhookRequestBodyUnmarshaller(
    V31WebhookRequestBodyValidator, WebhookRequestBodyUnmarshaller
):
    schema_unmarshallers_factory = oas31_schema_unmarshallers_factory


class V31WebhookRequestParametersUnmarshaller(
    V31WebhookRequestParametersValidator, WebhookRequestParametersUnmarshaller
):
    schema_unmarshallers_factory = oas31_schema_unmarshallers_factory


class V31WebhookRequestSecurityUnmarshaller(
    V31WebhookRequestSecurityValidator, WebhookRequestSecuritysUnmarshaller
):
    schema_unmarshallers_factory = oas31_schema_unmarshallers_factory


class V31WebhookRequestUnmarshaller(
    V31WebhookRequestValidator, WebhookRequestUnmarshaller
):
    schema_unmarshallers_factory = oas31_schema_unmarshallers_factory



from typing import Type
from typing import Union

from openapi_core.unmarshalling.response.protocols import ResponseUnmarshaller
from openapi_core.unmarshalling.response.protocols import (
    WebhookResponseUnmarshaller,
)

ResponseUnmarshallerType = Type[ResponseUnmarshaller]
WebhookResponseUnmarshallerType = Type[WebhookResponseUnmarshaller]
AnyResponseUnmarshallerType = Union[
    ResponseUnmarshallerType, WebhookResponseUnmarshallerType
]



"""OpenAPI core contrib falcon handlers module"""

from json import dumps
from typing import Any
from typing import Dict
from typing import Iterable
from typing import Type

from falcon import status_codes
from falcon.constants import MEDIA_JSON
from falcon.request import Request
from falcon.response import Response

from openapi_core.templating.media_types.exceptions import MediaTypeNotFound
from openapi_core.templating.paths.exceptions import OperationNotFound
from openapi_core.templating.paths.exceptions import PathNotFound
from openapi_core.templating.paths.exceptions import ServerNotFound
from openapi_core.templating.security.exceptions import SecurityNotFound
from openapi_core.unmarshalling.request.datatypes import RequestUnmarshalResult


class FalconOpenAPIErrorsHandler:
    OPENAPI_ERROR_STATUS: Dict[Type[BaseException], int] = {
        ServerNotFound: 400,
        SecurityNotFound: 403,
        OperationNotFound: 405,
        PathNotFound: 404,
        MediaTypeNotFound: 415,
    }

    def __init__(self, req: Request, resp: Response):
        self.req = req
        self.resp = resp

    def __call__(self, errors: Iterable[Exception]) -> Response:
        data_errors = [self.format_openapi_error(err) for err in errors]
        data = {
            "errors": data_errors,
        }
        data_str = dumps(data)
        data_error_max = max(data_errors, key=self.get_error_status)
        self.resp.content_type = MEDIA_JSON
        self.resp.status = getattr(
            status_codes,
            f"HTTP_{data_error_max['status']}",
            status_codes.HTTP_400,
        )
        self.resp.text = data_str
        self.resp.complete = True
        return self.resp

    @classmethod
    def format_openapi_error(cls, error: BaseException) -> Dict[str, Any]:
        if error.__cause__ is not None:
            error = error.__cause__
        return {
            "title": str(error),
            "status": cls.OPENAPI_ERROR_STATUS.get(error.__class__, 400),
            "type": str(type(error)),
        }

    @classmethod
    def get_error_status(cls, error: Dict[str, Any]) -> int:
        return int(error["status"])


class FalconOpenAPIValidRequestHandler:
    def __init__(self, req: Request, resp: Response):
        self.req = req
        self.resp = resp

    def __call__(
        self, request_unmarshal_result: RequestUnmarshalResult
    ) -> Response:
        self.req.context.openapi = request_unmarshal_result
        return self.resp



from dataclasses import dataclass
from dataclasses import field
from typing import Iterable

from openapi_core.exceptions import OpenAPIError


class ValidateError(OpenAPIError):
    """Schema validate operation error"""


@dataclass
class InvalidSchemaValue(ValidateError):
    """Value not valid for schema"""

    value: str
    type: str
    schema_errors: Iterable[Exception] = field(default_factory=list)

    def __str__(self) -> str:
        return (
            "Value {value} not valid for schema of type {type}: {errors}"
        ).format(value=self.value, type=self.type, errors=self.schema_errors)



"""OpenAPI core templating paths datatypes module"""

from collections import namedtuple

Path = namedtuple("Path", ["path", "path_result"])
PathOperation = namedtuple(
    "PathOperation", ["path", "operation", "path_result"]
)
PathOperationServer = namedtuple(
    "PathOperationServer",
    ["path", "operation", "server", "path_result", "server_result"],
)



from typing import Iterator
from typing import Optional
from typing import Protocol
from typing import runtime_checkable

from jsonschema_path import SchemaPath

from openapi_core.templating.paths.datatypes import Path
from openapi_core.templating.paths.datatypes import PathOperation
from openapi_core.templating.paths.datatypes import PathOperationServer


@runtime_checkable
class PathsIterator(Protocol):
    def __call__(
        self, name: str, spec: SchemaPath, base_url: Optional[str] = None
    ) -> Iterator[Path]: ...


@runtime_checkable
class OperationsIterator(Protocol):
    def __call__(
        self,
        method: str,
        paths_iter: Iterator[Path],
        spec: SchemaPath,
        base_url: Optional[str] = None,
    ) -> Iterator[PathOperation]: ...


@runtime_checkable
class ServersIterator(Protocol):
    def __call__(
        self,
        name: str,
        operations_iter: Iterator[PathOperation],
        spec: SchemaPath,
        base_url: Optional[str] = None,
    ) -> Iterator[PathOperationServer]: ...



from dataclasses import dataclass
from typing import Optional

from openapi_core.casting.schemas.factories import SchemaCastersFactory
from openapi_core.deserializing.media_types import (
    media_type_deserializers_factory,
)
from openapi_core.deserializing.media_types.datatypes import (
    MediaTypeDeserializersDict,
)
from openapi_core.deserializing.media_types.factories import (
    MediaTypeDeserializersFactory,
)
from openapi_core.deserializing.styles import style_deserializers_factory
from openapi_core.deserializing.styles.factories import (
    StyleDeserializersFactory,
)
from openapi_core.security import security_provider_factory
from openapi_core.security.factories import SecurityProviderFactory
from openapi_core.templating.paths.types import PathFinderType
from openapi_core.validation.schemas.datatypes import FormatValidatorsDict
from openapi_core.validation.schemas.factories import SchemaValidatorsFactory


@dataclass
class ValidatorConfig:
    """Validator configuration dataclass.

    Attributes:
        server_base_url
            Server base URI.
        path_finder_cls
            Path finder class.
        webhook_path_finder_cls
            Webhook path finder class.
        style_deserializers_factory
            Style deserializers factory.
        media_type_deserializers_factory
            Media type deserializers factory.
        schema_casters_factory
            Schema casters factory.
        schema_validators_factory
            Schema validators factory.
        extra_format_validators
            Extra format validators.
        extra_media_type_deserializers
            Extra media type deserializers.
        security_provider_factory
            Security providers factory.
    """

    server_base_url: Optional[str] = None
    path_finder_cls: Optional[PathFinderType] = None
    webhook_path_finder_cls: Optional[PathFinderType] = None

    style_deserializers_factory: StyleDeserializersFactory = (
        style_deserializers_factory
    )
    media_type_deserializers_factory: MediaTypeDeserializersFactory = (
        media_type_deserializers_factory
    )
    schema_casters_factory: Optional[SchemaCastersFactory] = None
    schema_validators_factory: Optional[SchemaValidatorsFactory] = None

    extra_format_validators: Optional[FormatValidatorsDict] = None
    extra_media_type_deserializers: Optional[MediaTypeDeserializersDict] = None

    security_provider_factory: SecurityProviderFactory = (
        security_provider_factory
    )



from typing import Optional

from fastapi import Response

from openapi_core.contrib.starlette.responses import StarletteOpenAPIResponse


class FastAPIOpenAPIResponse(StarletteOpenAPIResponse):
    def __init__(self, response: Response, data: Optional[bytes] = None):
        super().__init__(response, data=data)



from dataclasses import dataclass

from openapi_core.exceptions import OpenAPIError


class PathError(OpenAPIError):
    """Path error"""


@dataclass
class PathNotFound(PathError):
    """Path not found"""

    url: str

    def __str__(self) -> str:
        return f"Path not found for {self.url}"


@dataclass
class PathsNotFound(PathNotFound):
    """Paths not found"""

    def __str__(self) -> str:
        return f"Paths not found in spec: {self.url}"


@dataclass
class OperationNotFound(PathError):
    """Find path operation error"""

    url: str
    method: str

    def __str__(self) -> str:
        return f"Operation {self.method} not found for {self.url}"


@dataclass
class ServerNotFound(PathError):
    """Find server error"""

    url: str

    def __str__(self) -> str:
        return f"Server not found for {self.url}"



"""OpenAPI core contrib flask views module"""

from typing import Any

from flask.views import MethodView

from openapi_core import OpenAPI
from openapi_core.contrib.flask.decorators import FlaskOpenAPIViewDecorator
from openapi_core.contrib.flask.handlers import FlaskOpenAPIErrorsHandler


class FlaskOpenAPIView(MethodView):
    """Brings OpenAPI specification validation and unmarshalling for views."""

    openapi_errors_handler = FlaskOpenAPIErrorsHandler

    def __init__(self, openapi: OpenAPI):
        super().__init__()

        self.decorator = FlaskOpenAPIViewDecorator(
            openapi,
            errors_handler_cls=self.openapi_errors_handler,
        )

    def dispatch_request(self, *args: Any, **kwargs: Any) -> Any:
        response = self.decorator(super().dispatch_request)(*args, **kwargs)

        return response



"""OpenAPI core validation request datatypes module"""

from __future__ import annotations

from dataclasses import dataclass
from dataclasses import field
from typing import Any
from typing import Mapping

from werkzeug.datastructures import Headers
from werkzeug.datastructures import ImmutableMultiDict


@dataclass
class RequestParameters:
    """OpenAPI request parameters dataclass.

    Attributes:
        query
            Query string parameters as MultiDict. Must support getlist method.
        header
            Request headers as Headers.
        cookie
            Request cookies as MultiDict.
        path
            Path parameters as dict. Gets resolved against spec if empty.
    """

    query: Mapping[str, Any] = field(default_factory=ImmutableMultiDict)
    header: Mapping[str, Any] = field(default_factory=Headers)
    cookie: Mapping[str, Any] = field(default_factory=ImmutableMultiDict)
    path: Mapping[str, Any] = field(default_factory=dict)

    def __getitem__(self, location: str) -> Any:
        return getattr(self, location)


@dataclass
class Parameters:
    query: Mapping[str, Any] = field(default_factory=dict)
    header: Mapping[str, Any] = field(default_factory=dict)
    cookie: Mapping[str, Any] = field(default_factory=dict)
    path: Mapping[str, Any] = field(default_factory=dict)



from typing import Any
from typing import List
from typing import Protocol
from typing import runtime_checkable


@runtime_checkable
class SuportsGetAll(Protocol):
    def getall(self, name: str) -> List[Any]: ...


@runtime_checkable
class SuportsGetList(Protocol):
    def getlist(self, name: str) -> List[Any]: ...



from typing import Any
from typing import Dict

from jsonschema_path import SchemaPath


def is_absolute(url: str) -> bool:
    return url.startswith("//") or "://" in url


def get_server_default_variables(server: SchemaPath) -> Dict[str, Any]:
    if "variables" not in server:
        return {}

    defaults = {}
    variables = server / "variables"
    for name, variable in list(variables.items()):
        defaults[name] = variable["default"]
    return defaults


def get_server_url(server: SchemaPath, **variables: Any) -> str:
    if not variables:
        variables = get_server_default_variables(server)
    assert isinstance(server["url"], str)
    return server["url"].format(**variables)



"""OpenAPI core contrib flask decorators module"""

from functools import wraps
from typing import Any
from typing import Callable
from typing import Type

from flask.globals import request
from flask.wrappers import Request
from flask.wrappers import Response
from jsonschema_path import SchemaPath

from openapi_core import OpenAPI
from openapi_core.contrib.flask.handlers import FlaskOpenAPIErrorsHandler
from openapi_core.contrib.flask.handlers import FlaskOpenAPIValidRequestHandler
from openapi_core.contrib.flask.integrations import FlaskIntegration
from openapi_core.contrib.flask.providers import FlaskRequestProvider
from openapi_core.contrib.flask.requests import FlaskOpenAPIRequest
from openapi_core.contrib.flask.responses import FlaskOpenAPIResponse


class FlaskOpenAPIViewDecorator(FlaskIntegration):
    valid_request_handler_cls = FlaskOpenAPIValidRequestHandler
    errors_handler_cls: Type[FlaskOpenAPIErrorsHandler] = (
        FlaskOpenAPIErrorsHandler
    )

    def __init__(
        self,
        openapi: OpenAPI,
        request_cls: Type[FlaskOpenAPIRequest] = FlaskOpenAPIRequest,
        response_cls: Type[FlaskOpenAPIResponse] = FlaskOpenAPIResponse,
        request_provider: Type[FlaskRequestProvider] = FlaskRequestProvider,
        errors_handler_cls: Type[
            FlaskOpenAPIErrorsHandler
        ] = FlaskOpenAPIErrorsHandler,
    ):
        super().__init__(openapi)
        self.request_cls = request_cls
        self.response_cls = response_cls
        self.request_provider = request_provider
        self.errors_handler_cls = errors_handler_cls

    def __call__(self, view: Callable[..., Any]) -> Callable[..., Any]:
        @wraps(view)
        def decorated(*args: Any, **kwargs: Any) -> Response:
            request = self.get_request()
            valid_request_handler = self.valid_request_handler_cls(
                request, view, *args, **kwargs
            )
            errors_handler = self.errors_handler_cls()
            response = self.handle_request(
                request, valid_request_handler, errors_handler
            )
            return self.handle_response(request, response, errors_handler)

        return decorated

    def get_request(self) -> Request:
        return request

    @classmethod
    def from_spec(
        cls,
        spec: SchemaPath,
        request_cls: Type[FlaskOpenAPIRequest] = FlaskOpenAPIRequest,
        response_cls: Type[FlaskOpenAPIResponse] = FlaskOpenAPIResponse,
        request_provider: Type[FlaskRequestProvider] = FlaskRequestProvider,
        errors_handler_cls: Type[
            FlaskOpenAPIErrorsHandler
        ] = FlaskOpenAPIErrorsHandler,
    ) -> "FlaskOpenAPIViewDecorator":
        openapi = OpenAPI(spec)
        return cls(
            openapi,
            request_cls=request_cls,
            response_cls=response_cls,
            request_provider=request_provider,
            errors_handler_cls=errors_handler_cls,
        )



"""OpenAPI core validation response module"""

from typing import Mapping

from openapi_spec_validator.versions import consts as versions
from openapi_spec_validator.versions.datatypes import SpecVersion

from openapi_core.validation.response.types import ResponseValidatorType
from openapi_core.validation.response.types import WebhookResponseValidatorType
from openapi_core.validation.response.validators import (
    V30ResponseDataValidator,
)
from openapi_core.validation.response.validators import (
    V30ResponseHeadersValidator,
)
from openapi_core.validation.response.validators import V30ResponseValidator
from openapi_core.validation.response.validators import (
    V31ResponseDataValidator,
)
from openapi_core.validation.response.validators import (
    V31ResponseHeadersValidator,
)
from openapi_core.validation.response.validators import V31ResponseValidator
from openapi_core.validation.response.validators import (
    V31WebhookResponseDataValidator,
)
from openapi_core.validation.response.validators import (
    V31WebhookResponseHeadersValidator,
)
from openapi_core.validation.response.validators import (
    V31WebhookResponseValidator,
)

__all__ = [
    "VALIDATORS",
    "WEBHOOK_VALIDATORS",
    "V30ResponseDataValidator",
    "V30ResponseHeadersValidator",
    "V30ResponseValidator",
    "V31ResponseDataValidator",
    "V31ResponseHeadersValidator",
    "V31ResponseValidator",
    "V31WebhookResponseDataValidator",
    "V31WebhookResponseHeadersValidator",
    "V31WebhookResponseValidator",
    "V3ResponseValidator",
    "V3WebhookResponseValidator",
]

# versions mapping
VALIDATORS: Mapping[SpecVersion, ResponseValidatorType] = {
    versions.OPENAPIV30: V30ResponseValidator,
    versions.OPENAPIV31: V31ResponseValidator,
}
WEBHOOK_VALIDATORS: Mapping[SpecVersion, WebhookResponseValidatorType] = {
    versions.OPENAPIV31: V31WebhookResponseValidator,
}

# alias to the latest v3 version
V3ResponseValidator = V31ResponseValidator
V3WebhookResponseValidator = V31WebhookResponseValidator



from functools import wraps
from inspect import signature
from typing import Any
from typing import Callable
from typing import Optional
from typing import Type

from openapi_core.exceptions import OpenAPIError
from openapi_core.validation.schemas.exceptions import ValidateError

OpenAPIErrorType = Type[OpenAPIError]


class ValidationErrorWrapper:
    def __init__(
        self,
        err_cls: OpenAPIErrorType,
        err_validate_cls: Optional[OpenAPIErrorType] = None,
        err_cls_init: Optional[str] = None,
        **err_cls_kw: Any
    ):
        self.err_cls = err_cls
        self.err_validate_cls = err_validate_cls or err_cls
        self.err_cls_init = err_cls_init
        self.err_cls_kw = err_cls_kw

    def __call__(self, f: Callable[..., Any]) -> Callable[..., Any]:
        @wraps(f)
        def wrapper(*args: Any, **kwds: Any) -> Any:
            try:
                return f(*args, **kwds)
            except ValidateError as exc:
                self._raise_error(exc, self.err_validate_cls, f, *args, **kwds)
            except OpenAPIError as exc:
                self._raise_error(exc, self.err_cls, f, *args, **kwds)

        return wrapper

    def _raise_error(
        self,
        exc: OpenAPIError,
        cls: OpenAPIErrorType,
        f: Callable[..., Any],
        *args: Any,
        **kwds: Any
    ) -> None:
        if isinstance(exc, self.err_cls):
            raise
        sig = signature(f)
        ba = sig.bind(*args, **kwds)
        kw = {
            name: ba.arguments[func_kw]
            for name, func_kw in self.err_cls_kw.items()
        }
        init = cls
        if self.err_cls_init is not None:
            init = getattr(cls, self.err_cls_init)
        raise init(**kw) from exc



"""OpenAPI core templating media types finders module"""

import fnmatch
import re
from typing import Mapping
from typing import Tuple

from jsonschema_path import SchemaPath

from openapi_core.templating.media_types.datatypes import MediaType
from openapi_core.templating.media_types.exceptions import MediaTypeNotFound


class MediaTypeFinder:
    def __init__(self, content: SchemaPath):
        self.content = content

    def get_first(self) -> MediaType:
        mimetype, media_type = next(self.content.items())
        return MediaType(mimetype, {}, media_type)

    def find(self, mimetype: str) -> MediaType:
        if mimetype is None:
            raise MediaTypeNotFound(mimetype, list(self.content.keys()))

        mime_type, parameters = self._parse_mimetype(mimetype)

        # simple mime type
        for m in [mimetype, mime_type]:
            if m in self.content:
                return MediaType(mime_type, parameters, self.content / m)

        # range mime type
        if mime_type:
            for key, value in self.content.items():
                if fnmatch.fnmatch(mime_type, key):
                    return MediaType(key, parameters, value)

        raise MediaTypeNotFound(mimetype, list(self.content.keys()))

    def _parse_mimetype(self, mimetype: str) -> Tuple[str, Mapping[str, str]]:
        mimetype_parts = mimetype.split(";")
        mime_type = mimetype_parts[0].lower().rstrip()
        parameters = {}
        if len(mimetype_parts) > 1:
            parameters_list = (
                self._parse_parameter(param_str)
                for param_str in mimetype_parts[1:]
            )
            parameters = dict(parameters_list)
        return mime_type, parameters

    def _parse_parameter(self, parameter: str) -> Tuple[str, str]:
        """Parse a parameter according to RFC 9110.

        See https://www.rfc-editor.org/rfc/rfc9110.html#name-parameters

        Important points:
            * parameter names are case-insensitive
            * parameter values are case-sensitive
                except "charset" which is case-insensitive
                https://www.rfc-editor.org/rfc/rfc2046#section-4.1.2
        """
        name, value = parameter.split("=")
        name = name.lower().lstrip()
        # remove surrounding quotes from value
        value = re.sub('^"(.*)"$', r"\1", value, count=1)
        if name == "charset":
            value = value.lower()
        return name, value.rstrip()



from typing import Any
from typing import Dict
from typing import Generator
from typing import Tuple


def unpack_params(
    params: Dict[str, Any]
) -> Generator[Tuple[str, Any], None, None]:
    for k, v in params.items():
        if isinstance(v, list):
            for v2 in v:
                yield (k, v2)
        else:
            yield (k, v)



"""OpenAPI core app module"""

from functools import cached_property
from pathlib import Path
from typing import Optional

from jsonschema._utils import Unset
from jsonschema.validators import _UNSET
from jsonschema_path import SchemaPath
from jsonschema_path.handlers.protocols import SupportsRead
from jsonschema_path.typing import Schema
from openapi_spec_validator import validate
from openapi_spec_validator.validation.exceptions import ValidatorDetectError
from openapi_spec_validator.versions.datatypes import SpecVersion
from openapi_spec_validator.versions.exceptions import OpenAPIVersionNotFound
from openapi_spec_validator.versions.shortcuts import get_spec_version
from typing_extensions import Annotated
from typing_extensions import Doc

from openapi_core.configurations import Config
from openapi_core.exceptions import SpecError
from openapi_core.protocols import Request
from openapi_core.protocols import Response
from openapi_core.protocols import WebhookRequest
from openapi_core.types import AnyRequest
from openapi_core.unmarshalling.request import (
    UNMARSHALLERS as REQUEST_UNMARSHALLERS,
)
from openapi_core.unmarshalling.request import (
    WEBHOOK_UNMARSHALLERS as WEBHOOK_REQUEST_UNMARSHALLERS,
)
from openapi_core.unmarshalling.request.datatypes import RequestUnmarshalResult
from openapi_core.unmarshalling.request.protocols import RequestUnmarshaller
from openapi_core.unmarshalling.request.protocols import (
    WebhookRequestUnmarshaller,
)
from openapi_core.unmarshalling.request.types import RequestUnmarshallerType
from openapi_core.unmarshalling.request.types import (
    WebhookRequestUnmarshallerType,
)
from openapi_core.unmarshalling.response import (
    UNMARSHALLERS as RESPONSE_UNMARSHALLERS,
)
from openapi_core.unmarshalling.response import (
    WEBHOOK_UNMARSHALLERS as WEBHOOK_RESPONSE_UNMARSHALLERS,
)
from openapi_core.unmarshalling.response.datatypes import (
    ResponseUnmarshalResult,
)
from openapi_core.unmarshalling.response.protocols import ResponseUnmarshaller
from openapi_core.unmarshalling.response.protocols import (
    WebhookResponseUnmarshaller,
)
from openapi_core.unmarshalling.response.types import ResponseUnmarshallerType
from openapi_core.unmarshalling.response.types import (
    WebhookResponseUnmarshallerType,
)
from openapi_core.validation.request import VALIDATORS as REQUEST_VALIDATORS
from openapi_core.validation.request import (
    WEBHOOK_VALIDATORS as WEBHOOK_REQUEST_VALIDATORS,
)
from openapi_core.validation.request.protocols import RequestValidator
from openapi_core.validation.request.protocols import WebhookRequestValidator
from openapi_core.validation.request.types import RequestValidatorType
from openapi_core.validation.request.types import WebhookRequestValidatorType
from openapi_core.validation.response import VALIDATORS as RESPONSE_VALIDATORS
from openapi_core.validation.response import (
    WEBHOOK_VALIDATORS as WEBHOOK_RESPONSE_VALIDATORS,
)
from openapi_core.validation.response.protocols import ResponseValidator
from openapi_core.validation.response.protocols import WebhookResponseValidator
from openapi_core.validation.response.types import ResponseValidatorType
from openapi_core.validation.response.types import WebhookResponseValidatorType


class OpenAPI:
    """`OpenAPI` application class, the main entrypoint class for OpenAPI-core.

    OpenAPI can be created in multiple ways: from existing memory data or from storage such as local disk via ``from_*()`` APIs

    Read more information, in the
    [OpenAPI-core docs for First Steps](https://openapi-core.readthedocs.io/#first-steps).

    Examples:
        You can import the OpenAPI class directly from openapi_core:

        Create an OpenAPI from a dictionary:

        ```python
        from openapi_core import OpenAPI

        app = OpenAPI.from_dict(spec)
        ```

        Create an OpenAPI from a path object:

        ```python
        from openapi_core import OpenAPI

        app = OpenAPI.from_path(path)
        ```

        Create an OpenAPI from a file path:

        ```python
        from openapi_core import OpenAPI

        app = OpenAPI.from_file_path('spec.yaml')
        ```

        Create an OpenAPI from a file object:

        ```python
        from openapi_core import OpenAPI

        with open('spec.yaml') as f:
            app = OpenAPI.from_file(f)
        ```

    """

    def __init__(
        self,
        spec: Annotated[
            SchemaPath,
            Doc(
                """
                OpenAPI specification schema path object.
                """
            ),
        ],
        config: Annotated[
            Optional[Config],
            Doc(
                """
                Configuration object for the OpenAPI application.
                """
            ),
        ] = None,
    ):
        if not isinstance(spec, SchemaPath):
            raise TypeError("'spec' argument is not type of SchemaPath")

        self.spec = spec
        self.config = config or Config()

        self.check_spec()

    @classmethod
    def from_dict(
        cls,
        data: Annotated[
            Schema,
            Doc(
                """
                Dictionary representing the OpenAPI specification.
                """
            ),
        ],
        config: Annotated[
            Optional[Config],
            Doc(
                """
                Configuration object for the OpenAPI application.
                """
            ),
        ] = None,
        base_uri: Annotated[
            str,
            Doc(
                """
                Base URI for the OpenAPI specification.
                """
            ),
        ] = "",
    ) -> "OpenAPI":
        """Creates an `OpenAPI` from a dictionary.

        Example:
        ```python
        from openapi_core import OpenAPI

        app = OpenAPI.from_dict(spec)
        ```

        Returns:
            OpenAPI: An instance of the OpenAPI class.
        """
        sp = SchemaPath.from_dict(data, base_uri=base_uri)
        return cls(sp, config=config)

    @classmethod
    def from_path(
        cls,
        path: Annotated[
            Path,
            Doc(
                """
                Path object representing the OpenAPI specification file.
                """
            ),
        ],
        config: Annotated[
            Optional[Config],
            Doc(
                """
                Configuration object for the OpenAPI application.
                """
            ),
        ] = None,
    ) -> "OpenAPI":
        """Creates an `OpenAPI` from a [Path object](https://docs.python.org/3/library/pathlib.html#pathlib.Path).

        Example:
        ```python
        from openapi_core import OpenAPI

        app = OpenAPI.from_path(path)
        ```

        Returns:
            OpenAPI: An instance of the OpenAPI class.
        """
        sp = SchemaPath.from_path(path)
        return cls(sp, config=config)

    @classmethod
    def from_file_path(
        cls,
        file_path: Annotated[
            str,
            Doc(
                """
                File path string representing the OpenAPI specification file.
                """
            ),
        ],
        config: Annotated[
            Optional[Config],
            Doc(
                """
                Configuration object for the OpenAPI application.
                """
            ),
        ] = None,
    ) -> "OpenAPI":
        """Creates an `OpenAPI` from a file path string.

        Example:
        ```python
        from openapi_core import OpenAPI

        app = OpenAPI.from_file_path('spec.yaml')
        ```

        Returns:
            OpenAPI: An instance of the OpenAPI class.
        """
        sp = SchemaPath.from_file_path(file_path)
        return cls(sp, config=config)

    @classmethod
    def from_file(
        cls,
        fileobj: Annotated[
            SupportsRead,
            Doc(
                """
                File object representing the OpenAPI specification file.
                """
            ),
        ],
        config: Annotated[
            Optional[Config],
            Doc(
                """
                Configuration object for the OpenAPI application.
                """
            ),
        ] = None,
        base_uri: Annotated[
            str,
            Doc(
                """
                Base URI for the OpenAPI specification.
                """
            ),
        ] = "",
    ) -> "OpenAPI":
        """Creates an `OpenAPI` from a [file object](https://docs.python.org/3/glossary.html#term-file-object).

        Example:
        ```python
        from openapi_core import OpenAPI

        with open('spec.yaml') as f:
            app = OpenAPI.from_file(f)
        ```

        Returns:
            OpenAPI: An instance of the OpenAPI class.
        """
        sp = SchemaPath.from_file(fileobj, base_uri=base_uri)
        return cls(sp, config=config)

    def _get_version(self) -> SpecVersion:
        try:
            return get_spec_version(self.spec.contents())
        # backward compatibility
        except OpenAPIVersionNotFound:
            raise SpecError("Spec schema version not detected")

    def check_spec(self) -> None:
        if self.config.spec_validator_cls is None:
            return

        cls = None
        if self.config.spec_validator_cls is not _UNSET:
            cls = self.config.spec_validator_cls

        try:
            validate(
                self.spec.contents(),
                base_uri=self.config.spec_base_uri
                or self.spec.accessor.resolver._base_uri,  # type: ignore[attr-defined]
                cls=cls,
            )
        except ValidatorDetectError:
            raise SpecError("spec not detected")

    @property
    def version(self) -> SpecVersion:
        return self._get_version()

    @cached_property
    def request_validator_cls(self) -> Optional[RequestValidatorType]:
        if not isinstance(self.config.request_validator_cls, Unset):
            return self.config.request_validator_cls
        return REQUEST_VALIDATORS.get(self.version)

    @cached_property
    def response_validator_cls(self) -> Optional[ResponseValidatorType]:
        if not isinstance(self.config.response_validator_cls, Unset):
            return self.config.response_validator_cls
        return RESPONSE_VALIDATORS.get(self.version)

    @cached_property
    def webhook_request_validator_cls(
        self,
    ) -> Optional[WebhookRequestValidatorType]:
        if not isinstance(self.config.webhook_request_validator_cls, Unset):
            return self.config.webhook_request_validator_cls
        return WEBHOOK_REQUEST_VALIDATORS.get(self.version)

    @cached_property
    def webhook_response_validator_cls(
        self,
    ) -> Optional[WebhookResponseValidatorType]:
        if not isinstance(self.config.webhook_response_validator_cls, Unset):
            return self.config.webhook_response_validator_cls
        return WEBHOOK_RESPONSE_VALIDATORS.get(self.version)

    @cached_property
    def request_unmarshaller_cls(self) -> Optional[RequestUnmarshallerType]:
        if not isinstance(self.config.request_unmarshaller_cls, Unset):
            return self.config.request_unmarshaller_cls
        return REQUEST_UNMARSHALLERS.get(self.version)

    @cached_property
    def response_unmarshaller_cls(self) -> Optional[ResponseUnmarshallerType]:
        if not isinstance(self.config.response_unmarshaller_cls, Unset):
            return self.config.response_unmarshaller_cls
        return RESPONSE_UNMARSHALLERS.get(self.version)

    @cached_property
    def webhook_request_unmarshaller_cls(
        self,
    ) -> Optional[WebhookRequestUnmarshallerType]:
        if not isinstance(self.config.webhook_request_unmarshaller_cls, Unset):
            return self.config.webhook_request_unmarshaller_cls
        return WEBHOOK_REQUEST_UNMARSHALLERS.get(self.version)

    @cached_property
    def webhook_response_unmarshaller_cls(
        self,
    ) -> Optional[WebhookResponseUnmarshallerType]:
        if not isinstance(
            self.config.webhook_response_unmarshaller_cls, Unset
        ):
            return self.config.webhook_response_unmarshaller_cls
        return WEBHOOK_RESPONSE_UNMARSHALLERS.get(self.version)

    @cached_property
    def request_validator(self) -> RequestValidator:
        if self.request_validator_cls is None:
            raise SpecError("Validator class not found")
        return self.request_validator_cls(
            self.spec,
            base_url=self.config.server_base_url,
            style_deserializers_factory=self.config.style_deserializers_factory,
            media_type_deserializers_factory=self.config.media_type_deserializers_factory,
            schema_casters_factory=self.config.schema_casters_factory,
            schema_validators_factory=self.config.schema_validators_factory,
            path_finder_cls=self.config.path_finder_cls,
            spec_validator_cls=self.config.spec_validator_cls,
            extra_format_validators=self.config.extra_format_validators,
            extra_media_type_deserializers=self.config.extra_media_type_deserializers,
            security_provider_factory=self.config.security_provider_factory,
        )

    @cached_property
    def response_validator(self) -> ResponseValidator:
        if self.response_validator_cls is None:
            raise SpecError("Validator class not found")
        return self.response_validator_cls(
            self.spec,
            base_url=self.config.server_base_url,
            style_deserializers_factory=self.config.style_deserializers_factory,
            media_type_deserializers_factory=self.config.media_type_deserializers_factory,
            schema_casters_factory=self.config.schema_casters_factory,
            schema_validators_factory=self.config.schema_validators_factory,
            path_finder_cls=self.config.path_finder_cls,
            spec_validator_cls=self.config.spec_validator_cls,
            extra_format_validators=self.config.extra_format_validators,
            extra_media_type_deserializers=self.config.extra_media_type_deserializers,
        )

    @cached_property
    def webhook_request_validator(self) -> WebhookRequestValidator:
        if self.webhook_request_validator_cls is None:
            raise SpecError("Validator class not found")
        return self.webhook_request_validator_cls(
            self.spec,
            base_url=self.config.server_base_url,
            style_deserializers_factory=self.config.style_deserializers_factory,
            media_type_deserializers_factory=self.config.media_type_deserializers_factory,
            schema_casters_factory=self.config.schema_casters_factory,
            schema_validators_factory=self.config.schema_validators_factory,
            path_finder_cls=self.config.webhook_path_finder_cls,
            spec_validator_cls=self.config.spec_validator_cls,
            extra_format_validators=self.config.extra_format_validators,
            extra_media_type_deserializers=self.config.extra_media_type_deserializers,
            security_provider_factory=self.config.security_provider_factory,
        )

    @cached_property
    def webhook_response_validator(self) -> WebhookResponseValidator:
        if self.webhook_response_validator_cls is None:
            raise SpecError("Validator class not found")
        return self.webhook_response_validator_cls(
            self.spec,
            base_url=self.config.server_base_url,
            style_deserializers_factory=self.config.style_deserializers_factory,
            media_type_deserializers_factory=self.config.media_type_deserializers_factory,
            schema_casters_factory=self.config.schema_casters_factory,
            schema_validators_factory=self.config.schema_validators_factory,
            path_finder_cls=self.config.webhook_path_finder_cls,
            spec_validator_cls=self.config.spec_validator_cls,
            extra_format_validators=self.config.extra_format_validators,
            extra_media_type_deserializers=self.config.extra_media_type_deserializers,
        )

    @cached_property
    def request_unmarshaller(self) -> RequestUnmarshaller:
        if self.request_unmarshaller_cls is None:
            raise SpecError("Unmarshaller class not found")
        return self.request_unmarshaller_cls(
            self.spec,
            base_url=self.config.server_base_url,
            style_deserializers_factory=self.config.style_deserializers_factory,
            media_type_deserializers_factory=self.config.media_type_deserializers_factory,
            schema_casters_factory=self.config.schema_casters_factory,
            schema_validators_factory=self.config.schema_validators_factory,
            path_finder_cls=self.config.path_finder_cls,
            spec_validator_cls=self.config.spec_validator_cls,
            extra_format_validators=self.config.extra_format_validators,
            extra_media_type_deserializers=self.config.extra_media_type_deserializers,
            security_provider_factory=self.config.security_provider_factory,
            schema_unmarshallers_factory=self.config.schema_unmarshallers_factory,
            extra_format_unmarshallers=self.config.extra_format_unmarshallers,
        )

    @cached_property
    def response_unmarshaller(self) -> ResponseUnmarshaller:
        if self.response_unmarshaller_cls is None:
            raise SpecError("Unmarshaller class not found")
        return self.response_unmarshaller_cls(
            self.spec,
            base_url=self.config.server_base_url,
            style_deserializers_factory=self.config.style_deserializers_factory,
            media_type_deserializers_factory=self.config.media_type_deserializers_factory,
            schema_casters_factory=self.config.schema_casters_factory,
            schema_validators_factory=self.config.schema_validators_factory,
            path_finder_cls=self.config.path_finder_cls,
            spec_validator_cls=self.config.spec_validator_cls,
            extra_format_validators=self.config.extra_format_validators,
            extra_media_type_deserializers=self.config.extra_media_type_deserializers,
            schema_unmarshallers_factory=self.config.schema_unmarshallers_factory,
            extra_format_unmarshallers=self.config.extra_format_unmarshallers,
        )

    @cached_property
    def webhook_request_unmarshaller(self) -> WebhookRequestUnmarshaller:
        if self.webhook_request_unmarshaller_cls is None:
            raise SpecError("Unmarshaller class not found")
        return self.webhook_request_unmarshaller_cls(
            self.spec,
            base_url=self.config.server_base_url,
            style_deserializers_factory=self.config.style_deserializers_factory,
            media_type_deserializers_factory=self.config.media_type_deserializers_factory,
            schema_casters_factory=self.config.schema_casters_factory,
            schema_validators_factory=self.config.schema_validators_factory,
            path_finder_cls=self.config.webhook_path_finder_cls,
            spec_validator_cls=self.config.spec_validator_cls,
            extra_format_validators=self.config.extra_format_validators,
            extra_media_type_deserializers=self.config.extra_media_type_deserializers,
            security_provider_factory=self.config.security_provider_factory,
            schema_unmarshallers_factory=self.config.schema_unmarshallers_factory,
            extra_format_unmarshallers=self.config.extra_format_unmarshallers,
        )

    @cached_property
    def webhook_response_unmarshaller(self) -> WebhookResponseUnmarshaller:
        if self.webhook_response_unmarshaller_cls is None:
            raise SpecError("Unmarshaller class not found")
        return self.webhook_response_unmarshaller_cls(
            self.spec,
            base_url=self.config.server_base_url,
            style_deserializers_factory=self.config.style_deserializers_factory,
            media_type_deserializers_factory=self.config.media_type_deserializers_factory,
            schema_casters_factory=self.config.schema_casters_factory,
            schema_validators_factory=self.config.schema_validators_factory,
            path_finder_cls=self.config.webhook_path_finder_cls,
            spec_validator_cls=self.config.spec_validator_cls,
            extra_format_validators=self.config.extra_format_validators,
            extra_media_type_deserializers=self.config.extra_media_type_deserializers,
            schema_unmarshallers_factory=self.config.schema_unmarshallers_factory,
            extra_format_unmarshallers=self.config.extra_format_unmarshallers,
        )

    def validate_request(
        self,
        request: Annotated[
            AnyRequest,
            Doc(
                """
                Request object to be validated.
                """
            ),
        ],
    ) -> None:
        """Validates the given request object.

        Args:
            request (AnyRequest): Request object to be validated.

        Raises:
            TypeError: If the request object is not of the expected type.
            SpecError: If the validator class is not found.
        """
        if isinstance(request, WebhookRequest):
            self.validate_webhook_request(request)
        else:
            self.validate_apicall_request(request)

    def validate_response(
        self,
        request: Annotated[
            AnyRequest,
            Doc(
                """
                Request object associated with the response.
                """
            ),
        ],
        response: Annotated[
            Response,
            Doc(
                """
                Response object to be validated.
                """
            ),
        ],
    ) -> None:
        """Validates the given response object associated with the request.

        Args:
            request (AnyRequest): Request object associated with the response.
            response (Response): Response object to be validated.

        Raises:
            TypeError: If the request or response object is not of the expected type.
            SpecError: If the validator class is not found.
        """
        if isinstance(request, WebhookRequest):
            self.validate_webhook_response(request, response)
        else:
            self.validate_apicall_response(request, response)

    def validate_apicall_request(
        self,
        request: Annotated[
            Request,
            Doc(
                """
                API call request object to be validated.
                """
            ),
        ],
    ) -> None:
        if not isinstance(request, Request):
            raise TypeError("'request' argument is not type of Request")
        self.request_validator.validate(request)

    def validate_apicall_response(
        self,
        request: Annotated[
            Request,
            Doc(
                """
                API call request object associated with the response.
                """
            ),
        ],
        response: Annotated[
            Response,
            Doc(
                """
                API call response object to be validated.
                """
            ),
        ],
    ) -> None:
        if not isinstance(request, Request):
            raise TypeError("'request' argument is not type of Request")
        if not isinstance(response, Response):
            raise TypeError("'response' argument is not type of Response")
        self.response_validator.validate(request, response)

    def validate_webhook_request(
        self,
        request: Annotated[
            WebhookRequest,
            Doc(
                """
                Webhook request object to be validated.
                """
            ),
        ],
    ) -> None:
        if not isinstance(request, WebhookRequest):
            raise TypeError("'request' argument is not type of WebhookRequest")
        self.webhook_request_validator.validate(request)

    def validate_webhook_response(
        self,
        request: Annotated[
            WebhookRequest,
            Doc(
                """
                Webhook request object associated with the response.
                """
            ),
        ],
        response: Annotated[
            Response,
            Doc(
                """
                Webhook response object to be validated.
                """
            ),
        ],
    ) -> None:
        if not isinstance(request, WebhookRequest):
            raise TypeError("'request' argument is not type of WebhookRequest")
        if not isinstance(response, Response):
            raise TypeError("'response' argument is not type of Response")
        self.webhook_response_validator.validate(request, response)

    def unmarshal_request(
        self,
        request: Annotated[
            AnyRequest,
            Doc(
                """
                Request object to be unmarshalled.
                """
            ),
        ],
    ) -> RequestUnmarshalResult:
        """Unmarshals the given request object.

        Args:
            request (AnyRequest): Request object to be unmarshalled.

        Returns:
            RequestUnmarshalResult: The result of the unmarshalling process.

        Raises:
            TypeError: If the request object is not of the expected type.
            SpecError: If the unmarshaller class is not found.
        """
        if isinstance(request, WebhookRequest):
            return self.unmarshal_webhook_request(request)
        else:
            return self.unmarshal_apicall_request(request)

    def unmarshal_response(
        self,
        request: Annotated[
            AnyRequest,
            Doc(
                """
                Request object associated with the response.
                """
            ),
        ],
        response: Annotated[
            Response,
            Doc(
                """
                Response object to be unmarshalled.
                """
            ),
        ],
    ) -> ResponseUnmarshalResult:
        """Unmarshals the given response object associated with the request.

        Args:
            request (AnyRequest): Request object associated with the response.
            response (Response): Response object to be unmarshalled.

        Returns:
            ResponseUnmarshalResult: The result of the unmarshalling process.

        Raises:
            TypeError: If the request or response object is not of the expected type.
            SpecError: If the unmarshaller class is not found.
        """
        if isinstance(request, WebhookRequest):
            return self.unmarshal_webhook_response(request, response)
        else:
            return self.unmarshal_apicall_response(request, response)

    def unmarshal_apicall_request(
        self,
        request: Annotated[
            Request,
            Doc(
                """
                API call request object to be unmarshalled.
                """
            ),
        ],
    ) -> RequestUnmarshalResult:
        if not isinstance(request, Request):
            raise TypeError("'request' argument is not type of Request")
        return self.request_unmarshaller.unmarshal(request)

    def unmarshal_apicall_response(
        self,
        request: Annotated[
            Request,
            Doc(
                """
                API call request object associated with the response.
                """
            ),
        ],
        response: Annotated[
            Response,
            Doc(
                """
                API call response object to be unmarshalled.
                """
            ),
        ],
    ) -> ResponseUnmarshalResult:
        if not isinstance(request, Request):
            raise TypeError("'request' argument is not type of Request")
        if not isinstance(response, Response):
            raise TypeError("'response' argument is not type of Response")
        return self.response_unmarshaller.unmarshal(request, response)

    def unmarshal_webhook_request(
        self,
        request: Annotated[
            WebhookRequest,
            Doc(
                """
                Webhook request object to be unmarshalled.
                """
            ),
        ],
    ) -> RequestUnmarshalResult:
        if not isinstance(request, WebhookRequest):
            raise TypeError("'request' argument is not type of WebhookRequest")
        return self.webhook_request_unmarshaller.unmarshal(request)

    def unmarshal_webhook_response(
        self,
        request: Annotated[
            WebhookRequest,
            Doc(
                """
                Webhook request object associated with the response.
                """
            ),
        ],
        response: Annotated[
            Response,
            Doc(
                """
                Webhook response object to be unmarshalled.
                """
            ),
        ],
    ) -> ResponseUnmarshalResult:
        if not isinstance(request, WebhookRequest):
            raise TypeError("'request' argument is not type of WebhookRequest")
        if not isinstance(response, Response):
            raise TypeError("'response' argument is not type of Response")
        return self.webhook_response_unmarshaller.unmarshal(request, response)



"""OpenAPI core unmarshalling processors module"""

from typing import Generic

from openapi_core.app import OpenAPI
from openapi_core.protocols import Request
from openapi_core.protocols import Response
from openapi_core.typing import RequestType
from openapi_core.typing import ResponseType


class ValidationIntegration(Generic[RequestType, ResponseType]):
    def __init__(
        self,
        openapi: OpenAPI,
    ):
        self.openapi = openapi

    def get_openapi_request(self, request: RequestType) -> Request:
        raise NotImplementedError

    def get_openapi_response(self, response: ResponseType) -> Response:
        raise NotImplementedError

    def validate_request(self, request: RequestType) -> None:
        openapi_request = self.get_openapi_request(request)
        self.openapi.validate_request(
            openapi_request,
        )

    def validate_response(
        self,
        request: RequestType,
        response: ResponseType,
    ) -> None:
        openapi_request = self.get_openapi_request(request)
        openapi_response = self.get_openapi_response(response)
        self.openapi.validate_response(openapi_request, openapi_response)



from openapi_core.contrib.fastapi.middlewares import FastAPIOpenAPIMiddleware
from openapi_core.contrib.fastapi.requests import FastAPIOpenAPIRequest
from openapi_core.contrib.fastapi.responses import FastAPIOpenAPIResponse

__all__ = [
    "FastAPIOpenAPIMiddleware",
    "FastAPIOpenAPIRequest",
    "FastAPIOpenAPIResponse",
]



"""OpenAPI core validation validators module"""

import warnings
from functools import cached_property
from typing import Any
from typing import Mapping
from typing import Optional
from typing import Tuple
from urllib.parse import urljoin

from jsonschema_path import SchemaPath
from openapi_spec_validator.validation.types import SpecValidatorType

from openapi_core.casting.schemas.factories import SchemaCastersFactory
from openapi_core.deserializing.media_types import (
    media_type_deserializers_factory,
)
from openapi_core.deserializing.media_types.datatypes import (
    MediaTypeDeserializersDict,
)
from openapi_core.deserializing.media_types.factories import (
    MediaTypeDeserializersFactory,
)
from openapi_core.deserializing.styles import style_deserializers_factory
from openapi_core.deserializing.styles.exceptions import (
    EmptyQueryParameterValue,
)
from openapi_core.deserializing.styles.factories import (
    StyleDeserializersFactory,
)
from openapi_core.protocols import Request
from openapi_core.protocols import WebhookRequest
from openapi_core.schema.parameters import get_style_and_explode
from openapi_core.templating.media_types.datatypes import MediaType
from openapi_core.templating.paths.datatypes import PathOperationServer
from openapi_core.templating.paths.finders import APICallPathFinder
from openapi_core.templating.paths.finders import BasePathFinder
from openapi_core.templating.paths.finders import WebhookPathFinder
from openapi_core.templating.paths.types import PathFinderType
from openapi_core.validation.schemas.datatypes import FormatValidatorsDict
from openapi_core.validation.schemas.factories import SchemaValidatorsFactory


class BaseValidator:
    schema_casters_factory: SchemaCastersFactory = NotImplemented
    schema_validators_factory: SchemaValidatorsFactory = NotImplemented
    path_finder_cls: PathFinderType = NotImplemented
    spec_validator_cls: Optional[SpecValidatorType] = None

    def __init__(
        self,
        spec: SchemaPath,
        base_url: Optional[str] = None,
        style_deserializers_factory: StyleDeserializersFactory = style_deserializers_factory,
        media_type_deserializers_factory: MediaTypeDeserializersFactory = media_type_deserializers_factory,
        schema_casters_factory: Optional[SchemaCastersFactory] = None,
        schema_validators_factory: Optional[SchemaValidatorsFactory] = None,
        path_finder_cls: Optional[PathFinderType] = None,
        spec_validator_cls: Optional[SpecValidatorType] = None,
        format_validators: Optional[FormatValidatorsDict] = None,
        extra_format_validators: Optional[FormatValidatorsDict] = None,
        extra_media_type_deserializers: Optional[
            MediaTypeDeserializersDict
        ] = None,
    ):
        self.spec = spec
        self.base_url = base_url

        self.schema_casters_factory = (
            schema_casters_factory or self.schema_casters_factory
        )
        if self.schema_casters_factory is NotImplemented:
            raise NotImplementedError("schema_casters_factory is not assigned")
        self.style_deserializers_factory = style_deserializers_factory
        self.media_type_deserializers_factory = (
            media_type_deserializers_factory
        )
        self.schema_validators_factory = (
            schema_validators_factory or self.schema_validators_factory
        )
        if self.schema_validators_factory is NotImplemented:
            raise NotImplementedError(
                "schema_validators_factory is not assigned"
            )
        self.path_finder_cls = path_finder_cls or self.path_finder_cls
        if self.path_finder_cls is NotImplemented:
            raise NotImplementedError("path_finder_cls is not assigned")
        self.spec_validator_cls = spec_validator_cls or self.spec_validator_cls
        self.format_validators = format_validators
        self.extra_format_validators = extra_format_validators
        self.extra_media_type_deserializers = extra_media_type_deserializers

    @cached_property
    def path_finder(self) -> BasePathFinder:
        return self.path_finder_cls(self.spec, base_url=self.base_url)

    def check_spec(self, spec: SchemaPath) -> None:
        if self.spec_validator_cls is None:
            return

        validator = self.spec_validator_cls(spec)
        validator.validate()

    def _find_media_type(
        self, content: SchemaPath, mimetype: Optional[str] = None
    ) -> MediaType:
        from openapi_core.templating.media_types.finders import MediaTypeFinder

        finder = MediaTypeFinder(content)
        if mimetype is None:
            return finder.get_first()
        return finder.find(mimetype)

    def _deserialise_media_type(
        self,
        media_type: SchemaPath,
        mimetype: str,
        parameters: Mapping[str, str],
        value: bytes,
    ) -> Any:
        schema = media_type.get("schema")
        encoding = None
        if "encoding" in media_type:
            encoding = media_type.get("encoding")
        deserializer = self.media_type_deserializers_factory.create(
            mimetype,
            schema=schema,
            parameters=parameters,
            encoding=encoding,
            extra_media_type_deserializers=self.extra_media_type_deserializers,
        )
        return deserializer.deserialize(value)

    def _deserialise_style(
        self,
        param_or_header: SchemaPath,
        location: Mapping[str, Any],
        name: Optional[str] = None,
    ) -> Any:
        name = name or param_or_header["name"]
        style, explode = get_style_and_explode(param_or_header)
        schema = param_or_header / "schema"
        deserializer = self.style_deserializers_factory.create(
            style, explode, schema, name=name
        )
        return deserializer.deserialize(location)

    def _cast(self, schema: SchemaPath, value: Any) -> Any:
        caster = self.schema_casters_factory.create(schema)
        return caster.cast(value)

    def _validate_schema(self, schema: SchemaPath, value: Any) -> None:
        validator = self.schema_validators_factory.create(
            schema,
            format_validators=self.format_validators,
            extra_format_validators=self.extra_format_validators,
        )
        validator.validate(value)

    def _get_param_or_header_and_schema(
        self,
        param_or_header: SchemaPath,
        location: Mapping[str, Any],
        name: Optional[str] = None,
    ) -> Tuple[Any, Optional[SchemaPath]]:
        schema: Optional[SchemaPath] = None
        # Simple scenario
        if "content" not in param_or_header:
            casted, schema = self._get_simple_param_or_header(
                param_or_header, location, name=name
            )
        # Complex scenario
        else:
            casted, schema = self._get_complex_param_or_header(
                param_or_header, location, name=name
            )

        if schema is None:
            return casted, None
        self._validate_schema(schema, casted)
        return casted, schema

    def _get_simple_param_or_header(
        self,
        param_or_header: SchemaPath,
        location: Mapping[str, Any],
        name: Optional[str] = None,
    ) -> Tuple[Any, SchemaPath]:
        allow_empty_values = param_or_header.getkey("allowEmptyValue")
        if allow_empty_values:
            warnings.warn(
                "Use of allowEmptyValue property is deprecated",
                DeprecationWarning,
            )
        # in simple scenrios schema always exist
        schema = param_or_header / "schema"
        try:
            deserialised = self._deserialise_style(
                param_or_header, location, name=name
            )
        except KeyError:
            if "default" not in schema:
                raise
            return schema["default"], schema
        if allow_empty_values is not None:
            warnings.warn(
                "Use of allowEmptyValue property is deprecated",
                DeprecationWarning,
            )
        if allow_empty_values is None or not allow_empty_values:
            # if "in" not defined then it's a Header
            location_name = param_or_header.getkey("in", "header")
            if (
                location_name == "query"
                and deserialised == ""
                and not allow_empty_values
            ):
                param_or_header_name = param_or_header["name"]
                raise EmptyQueryParameterValue(param_or_header_name)
        casted = self._cast(schema, deserialised)
        return casted, schema

    def _get_complex_param_or_header(
        self,
        param_or_header: SchemaPath,
        location: Mapping[str, Any],
        name: Optional[str] = None,
    ) -> Tuple[Any, Optional[SchemaPath]]:
        content = param_or_header / "content"
        raw = self._get_media_type_value(param_or_header, location, name=name)
        return self._get_content_schema_value_and_schema(raw, content)

    def _get_content_schema_value_and_schema(
        self,
        raw: bytes,
        content: SchemaPath,
        mimetype: Optional[str] = None,
    ) -> Tuple[Any, Optional[SchemaPath]]:
        mime_type, parameters, media_type = self._find_media_type(
            content, mimetype
        )
        # no point to catch KetError
        # in complex scenrios schema doesn't exist
        deserialised = self._deserialise_media_type(
            media_type, mime_type, parameters, raw
        )

        if "schema" not in media_type:
            return deserialised, None

        schema = media_type / "schema"
        # cast for urlencoded content
        # FIXME: don't cast data from media type deserializer
        # See https://github.com/python-openapi/openapi-core/issues/706
        casted = self._cast(schema, deserialised)
        return casted, schema

    def _get_content_and_schema(
        self, raw: bytes, content: SchemaPath, mimetype: Optional[str] = None
    ) -> Tuple[Any, Optional[SchemaPath]]:
        casted, schema = self._get_content_schema_value_and_schema(
            raw, content, mimetype
        )
        if schema is None:
            return casted, None
        self._validate_schema(schema, casted)
        return casted, schema

    def _get_media_type_value(
        self,
        param_or_header: SchemaPath,
        location: Mapping[str, Any],
        name: Optional[str] = None,
    ) -> Any:
        name = name or param_or_header["name"]
        return location[name]


class BaseAPICallValidator(BaseValidator):
    path_finder_cls = APICallPathFinder

    def _find_path(self, request: Request) -> PathOperationServer:
        path_pattern = getattr(request, "path_pattern", None) or request.path
        full_url = urljoin(request.host_url, path_pattern)
        return self.path_finder.find(request.method, full_url)


class BaseWebhookValidator(BaseValidator):
    path_finder_cls = WebhookPathFinder

    def _find_path(self, request: WebhookRequest) -> PathOperationServer:
        return self.path_finder.find(request.method, request.name)



from typing import Any
from typing import Dict

from jsonschema_path import SchemaPath


def get_properties(schema: SchemaPath) -> Dict[str, Any]:
    properties = schema.get("properties", {})
    properties_dict = dict(list(properties.items()))
    return properties_dict



from typing import Tuple

from jsonschema_path import SchemaPath


def get_style(
    param_or_header: SchemaPath, default_location: str = "header"
) -> str:
    """Checks parameter/header style for simpler scenarios"""
    if "style" in param_or_header:
        assert isinstance(param_or_header["style"], str)
        return param_or_header["style"]

    location = param_or_header.getkey("in", default_location)

    # determine default
    return "simple" if location in ["path", "header"] else "form"


def get_explode(param_or_header: SchemaPath) -> bool:
    """Checks parameter/header explode for simpler scenarios"""
    if "explode" in param_or_header:
        assert isinstance(param_or_header["explode"], bool)
        return param_or_header["explode"]

    # determine default
    style = get_style(param_or_header)
    return style == "form"


def get_style_and_explode(
    param_or_header: SchemaPath, default_location: str = "header"
) -> Tuple[str, bool]:
    """Checks parameter/header explode for simpler scenarios"""
    style = get_style(param_or_header, default_location=default_location)
    if "explode" in param_or_header:
        assert isinstance(param_or_header["explode"], bool)
        return style, param_or_header["explode"]

    return style, style == "form"



from collections import OrderedDict

from openapi_core.casting.schemas.casters import ArrayCaster
from openapi_core.casting.schemas.casters import BooleanCaster
from openapi_core.casting.schemas.casters import IntegerCaster
from openapi_core.casting.schemas.casters import NumberCaster
from openapi_core.casting.schemas.casters import ObjectCaster
from openapi_core.casting.schemas.casters import PrimitiveCaster
from openapi_core.casting.schemas.casters import TypesCaster
from openapi_core.casting.schemas.factories import SchemaCastersFactory
from openapi_core.validation.schemas import (
    oas30_read_schema_validators_factory,
)
from openapi_core.validation.schemas import (
    oas30_write_schema_validators_factory,
)
from openapi_core.validation.schemas import oas31_schema_validators_factory

__all__ = [
    "oas30_write_schema_casters_factory",
    "oas30_read_schema_casters_factory",
    "oas31_schema_casters_factory",
]

oas30_casters_dict = OrderedDict(
    [
        ("object", ObjectCaster),
        ("array", ArrayCaster),
        ("boolean", BooleanCaster),
        ("integer", IntegerCaster),
        ("number", NumberCaster),
        ("string", PrimitiveCaster),
    ]
)
oas31_casters_dict = oas30_casters_dict.copy()
oas31_casters_dict.update(
    {
        "null": PrimitiveCaster,
    }
)

oas30_types_caster = TypesCaster(
    oas30_casters_dict,
    PrimitiveCaster,
)
oas31_types_caster = TypesCaster(
    oas31_casters_dict,
    PrimitiveCaster,
    multi=PrimitiveCaster,
)

oas30_write_schema_casters_factory = SchemaCastersFactory(
    oas30_write_schema_validators_factory,
    oas30_types_caster,
)

oas30_read_schema_casters_factory = SchemaCastersFactory(
    oas30_read_schema_validators_factory,
    oas30_types_caster,
)

oas31_schema_casters_factory = SchemaCastersFactory(
    oas31_schema_validators_factory,
    oas31_types_caster,
)
